# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import functools

from alembic import context
from alembic import op
import sqlalchemy as sa


def skip_if_offline(func):
    """Decorator for skipping migrations in offline mode."""
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        if context.is_offline_mode():
            return
        return func(*args, **kwargs)

    return decorator


def raise_if_offline(func):
    """Decorator for raising if a function is called in offline mode."""
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        if context.is_offline_mode():
            raise RuntimeError(_("%s cannot be called while in offline mode") %
                               func.__name__)
        return func(*args, **kwargs)

    return decorator


@raise_if_offline
def schema_has_table(table_name):
    """Check whether the specified table exists in the current schema.

    This method cannot be executed in offline mode.
    """
    bind = op.get_bind()
    insp = sa.engine.reflection.Inspector.from_engine(bind)
    return table_name in insp.get_table_names()


@raise_if_offline
def schema_has_column(table_name, column_name):
    """Check whether the specified column exists in the current schema.

    This method cannot be executed in offline mode.
    """
    bind = op.get_bind()
    insp = sa.engine.reflection.Inspector.from_engine(bind)
    # first check that the table exists
    if not schema_has_table(table_name):
        return
    # check whether column_name exists in table columns
    return column_name in [column['name'] for column in
                           insp.get_columns(table_name)]


@raise_if_offline
def alter_column_if_exists(table_name, column_name, **kwargs):
    """Alter a column only if it exists in the schema."""
    if schema_has_column(table_name, column_name):
        op.alter_column(table_name, column_name, **kwargs)


@raise_if_offline
def drop_table_if_exists(table_name):
    if schema_has_table(table_name):
        op.drop_table(table_name)


@raise_if_offline
def rename_table_if_exists(old_table_name, new_table_name):
    if schema_has_table(old_table_name):
        op.rename_table(old_table_name, new_table_name)


def alter_enum(table, column, enum_type, nullable):
    bind = op.get_bind()
    engine = bind.engine
    if engine.name == 'postgresql':
        values = {'table': table,
                  'column': column,
                  'name': enum_type.name}
        op.execute("ALTER TYPE %(name)s RENAME TO old_%(name)s" % values)
        enum_type.create(bind, checkfirst=False)
        op.execute("ALTER TABLE %(table)s RENAME COLUMN %(column)s TO "
                   "old_%(column)s" % values)
        op.add_column(table, sa.Column(column, enum_type, nullable=nullable))
        op.execute("UPDATE %(table)s SET %(column)s = "
                   "old_%(column)s::text::%(name)s" % values)
        op.execute("ALTER TABLE %(table)s DROP COLUMN old_%(column)s" % values)
        op.execute("DROP TYPE old_%(name)s" % values)
    else:
        op.alter_column(table, column, type_=enum_type,
                        existing_nullable=nullable)


def create_table_if_not_exist_psql(table_name, values):
    if op.get_bind().engine.dialect.server_version_info < (9, 1, 0):
        op.execute("CREATE LANGUAGE plpgsql")
    op.execute("CREATE OR REPLACE FUNCTION execute(TEXT) RETURNS VOID AS $$"
               "BEGIN EXECUTE $1; END;"
               "$$ LANGUAGE plpgsql STRICT;")
    op.execute("CREATE OR REPLACE FUNCTION table_exist(TEXT) RETURNS bool as "
               "$$ SELECT exists(select 1 from pg_class where relname=$1);"
               "$$ language sql STRICT;")
    op.execute("SELECT execute($$CREATE TABLE %(name)s %(columns)s $$) "
               "WHERE NOT table_exist(%(name)r);" %
               {'name': table_name,
                'columns': values})



# Copyright (c) 2014 Montavista Software, LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.datasources import cinder_driver
from congress.tests import base
from congress.tests.datasources import util
from congress.tests import helper

ResponseObj = util.ResponseObj


class TestCinderDriver(base.TestCase):

    def setUp(self):
        super(TestCinderDriver, self).setUp()
        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        self.driver = cinder_driver.CinderDriver(name='testcinder', args=args)

    def test_list_volumes(self):
        volumes_data = [
            ResponseObj({'id': '8bf2eddb-0e1a-46f9-a49a-853f8016f476',
                         'size': '1',
                         'user_id': 'b75055d5f0834d99ae874f085cf95272',
                         'status': 'available',
                         'description': 'foo',
                         'name': 'bar',
                         'bootable': 'False',
                         'created_at': '2014-10-09T12:16:23.000000',
                         'volume_type': 'lvmdriver-1'}),
            ResponseObj({'id': '7cd8f73d-3243-49c9-a25b-a77ceb6ad1fa',
                         'size': '1',
                         'user_id': '6e14edb203a84aa6a5a6a90872cbae79',
                         'status': 'creating',
                         'description': 'wonder',
                         'name': 'alice',
                         'bootable': 'True',
                         'created_at': '2014-10-12T06:54:55.000000',
                         'volume_type': 'None'})]

        volume_list = self.driver._translate_volumes(volumes_data)
        self.assertIsNotNone(volume_list)
        self.assertEqual(2, len(volume_list))

        self.assertEqual({('8bf2eddb-0e1a-46f9-a49a-853f8016f476', '1',
                           'b75055d5f0834d99ae874f085cf95272', 'available',
                           'foo', 'bar', 'False', '2014-10-09T12:16:23.000000',
                           'lvmdriver-1'),
                          ('7cd8f73d-3243-49c9-a25b-a77ceb6ad1fa', '1',
                           '6e14edb203a84aa6a5a6a90872cbae79', 'creating',
                           'wonder', 'alice', 'True',
                           '2014-10-12T06:54:55.000000', 'None')},
                         self.driver.state['volumes'])

    def test_list_snaphosts(self):
        snapshots_data = [
            ResponseObj({'status': 'available',
                         'created_at': '2014-10-12T06:54:55.000000',
                         'volume_id': 'b75055d5f0834d99ae874f085cf95272',
                         'size': '1',
                         'id': '7cd8f73d-3243-49c9-a25b-a77ceb6ad1fa',
                         'name': 'foo'}),
            ResponseObj({'status': 'creating',
                         'created_at': '2014-10-12T06:54:55.000000',
                         'volume_id': '6e14edb203a84aa6a5a6a90872cbae79',
                         'size': '1',
                         'id': '7cd8f73d-3243-49c9-a25b-a77ceb6ad1fa',
                         'name': 'baar'})]

        snapshot_list = self.driver._translate_snapshots(snapshots_data)
        self.assertIsNotNone(snapshot_list)
        self.assertEqual(2, len(snapshot_list))

        self.assertEqual({('7cd8f73d-3243-49c9-a25b-a77ceb6ad1fa', '1',
                           'available', 'b75055d5f0834d99ae874f085cf95272',
                           'foo', '2014-10-12T06:54:55.000000'),
                          ('7cd8f73d-3243-49c9-a25b-a77ceb6ad1fa', '1',
                           'creating', '6e14edb203a84aa6a5a6a90872cbae79',
                           'baar', '2014-10-12T06:54:55.000000')},
                         self.driver.state['snapshots'])

    def test_list_services(self):
        services_data = [
            ResponseObj({'status': 'enabled',
                         'binary': 'cinder-scheduler',
                         'zone': 'nova',
                         'state': 'up',
                         'updated_at': '2014-10-10T06:25:08.000000',
                         'host': 'openstack@lvmdriver-1',
                         'disabled_reason': 'None'}),
            ResponseObj({'status': 'enabled',
                         'binary': 'cinder-scheduler',
                         'zone': 'nova',
                         'state': 'up',
                         'updated_at': '2014-10-10T06:25:08.000000',
                         'host': 'openstack',
                         'disabled_reason': 'None'})]

        service_list = self.driver._translate_services(services_data)
        self.assertIsNotNone(service_list)
        self.assertEqual(2, len(service_list))

        self.assertEqual({('enabled', 'cinder-scheduler', 'nova',
                           'up', '2014-10-10T06:25:08.000000',
                           'openstack@lvmdriver-1', 'None'),
                          ('enabled', 'cinder-scheduler', 'nova',
                           'up', '2014-10-10T06:25:08.000000',
                           'openstack', 'None')},
                         self.driver.state['services'])

    def test_execute(self):
        class CinderClient(object):
            def __init__(self):
                self.testkey = None

            def createVolume(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        cinder_client = CinderClient()
        self.driver.cinder_client = cinder_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('createVolume', api_args)

        self.assertEqual(expected_ans, cinder_client.testkey)



# Copyright 2014 Plexxi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class dataObject(object):
    def __init__(self, data=None, version=0):

        if data is None:
            self.data = {}
        else:
            self.data = data

        if version:
            self.version = version
        else:
            self.version = int(bool(data))

    def __str__(self):
        return str(self.data)


class subData(object):
    """A piece of data that a data service is subscribed to.

    Each data service in the cage can have its own instance of
    this data; keep track of who published which instance.
    """
    def __init__(self, key, dataindex, corrId, callback):
        self.key = key
        self.dataindex = dataindex
        self.corrId = corrId
        self.callback = callback
        self.dataObjects = {}
        # LOG.info(
        #     "*****New subdata: %s, %s, %s",
        #     key, dataindex, id(self.dataObjects))

    def getSources(self):
        return self.dataObjects.keys()

    def update(self, sender, newdata):
        self.dataObjects[sender] = newdata

    def version(self, sender):
        version = 0

        if sender in self.dataObjects:
            version = self.dataObjects[sender].version

        return version

    def getData(self, sender):
        result = dataObject()

        if sender in self.dataObjects:
            LOG.info("subdata object: %s", self.dataObjects[sender])
            result = self.dataObjects[sender]

        return result

    def getAllData(self):
        result = {}
        for sender in self.dataObjects:
            result[sender] = self.dataObjects[sender]

        return result


class pubData(object):
    """A piece of data that a data service is publishing.

    Keep track of those data services that are subscribed.
    """
    def __init__(self, dataindex, args={}):
        self.dataindex = dataindex
        self.dataObject = dataObject()
        self.subscribers = {}
        self.requesters = {}
        self.args = args

    def update(self, newdata):
        version = self.dataObject.version + 1
        self.dataObject = dataObject(newdata, version)

    def get(self):
        return self.dataObject

    def version(self):
        return self.dataObject.version

    def addsubscriber(self, sender, type, corrId):
        if sender not in self.subscribers:
            self.subscribers[sender] = {}
            self.subscribers[sender]['type'] = type
            self.subscribers[sender]['correlationId'] = corrId

    def removesubscriber(self, sender):
        if sender in self.subscribers:
            del self.subscribers[sender]

    def getsubscribers(self, sender=""):
        if sender:
            if sender in self.subscribers:
                return self.subscribers[sender]
            else:
                return []
        else:
            return self.subscribers



#!/usr/bin/env python
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from cloudfoundryclient.v2 import client
from oslo_log import log as logging

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return CloudFoundryV2Driver(name, keys, inbox, datapath, args)


class CloudFoundryV2Driver(datasource_driver.PollingDataSourceDriver,
                           datasource_driver.ExecutionDriver):
    ORGANIZATIONS = 'organizations'
    SERVICE_BINDINGS = 'service_bindings'
    APPS = 'apps'
    SPACES = 'spaces'
    SERVICES = 'services'

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    organizations_translator = {
        'translation-type': 'HDICT',
        'table-name': ORGANIZATIONS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'guid', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans})}

    service_bindings_translator = {
        'translation-type': 'LIST',
        'table-name': SERVICE_BINDINGS,
        'parent-key': 'guid',
        'parent-col-name': 'app_guid',
        'val-col': 'service_instance_guid',
        'translator': value_trans}

    apps_translator = {
        'translation-type': 'HDICT',
        'table-name': APPS,
        'in-list': True,
        'parent-key': 'guid',
        'parent-col-name': 'space_guid',
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'guid', 'translator': value_trans},
             {'fieldname': 'buildpack', 'translator': value_trans},
             {'fieldname': 'command', 'translator': value_trans},
             {'fieldname': 'console', 'translator': value_trans},
             {'fieldname': 'debug', 'translator': value_trans},
             {'fieldname': 'detected_buildpack', 'translator': value_trans},
             {'fieldname': 'detected_start_command',
              'translator': value_trans},
             {'fieldname': 'disk_quota', 'translator': value_trans},
             {'fieldname': 'docker_image', 'translator': value_trans},
             {'fieldname': 'environment_json', 'translator': value_trans},
             {'fieldname': 'health_check_timeout', 'translator': value_trans},
             {'fieldname': 'instances', 'translator': value_trans},
             {'fieldname': 'memory', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'package_state', 'translator': value_trans},
             {'fieldname': 'package_updated_at', 'translator': value_trans},
             {'fieldname': 'production', 'translator': value_trans},
             {'fieldname': 'staging_failed_reason', 'translator': value_trans},
             {'fieldname': 'staging_task_id', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'version', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'service_bindings',
              'translator': service_bindings_translator})}

    spaces_translator = {
        'translation-type': 'HDICT',
        'table-name': SPACES,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'guid', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'apps', 'translator': apps_translator})}

    services_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVICES,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'guid', 'translator': value_trans},
             {'fieldname': 'space_guid', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'bound_app_count', 'translator': value_trans},
             {'fieldname': 'last_operation', 'translator': value_trans},
             {'fieldname': 'service_plan_name', 'translator': value_trans})}

    TRANSLATORS = [organizations_translator,
                   spaces_translator, services_translator]

    def __init__(self, name='', keys='', inbox=None,
                 datapath=None, args=None):
        super(CloudFoundryV2Driver, self).__init__(name, keys, inbox,
                                                   datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        self.cloudfoundry = client.Client(username=self.creds['username'],
                                          password=self.creds['password'],
                                          base_url=self.creds['auth_url'])
        self.cloudfoundry.login()
        self._cached_organizations = []
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'cloudfoundryv2'
        result['description'] = ('Datasource driver that interfaces with '
                                 'cloudfoundry')
        result['config'] = {'username': constants.REQUIRED,
                            'password': constants.REQUIRED,
                            'poll_time': constants.OPTIONAL,
                            'auth_url': constants.REQUIRED}
        result['secret'] = ['password']
        return result

    def _save_organizations(self, organizations):
        temp_organizations = []
        for organization in organizations['resources']:
            temp_organizations.append(organization['metadata']['guid'])
        self._cached_organizations = temp_organizations

    def _parse_services(self, services):
        data = []
        space_guid = services['guid']
        for service in services['services']:
            data.append(
                {'bound_app_count': service['bound_app_count'],
                 'guid': service['guid'],
                 'name': service['name'],
                 'service_plan_name': service['service_plan']['name'],
                 'space_guid': space_guid})
        return data

    def _get_app_services_guids(self, service_bindings):
        result = []
        for service_binding in service_bindings['resources']:
            result.append(service_binding['entity']['service_instance_guid'])
        return result

    def update_from_datasource(self):
        LOG.debug("CloudFoundry grabbing Data")
        organizations = self.cloudfoundry.get_organizations()
        self._translate_organizations(organizations)
        self._save_organizations(organizations)

        spaces = self._get_spaces()
        services = self._get_services_update_spaces(spaces)

        self._translate_spaces(spaces)
        self._translate_services(services)

    def _get_services_update_spaces(self, spaces):
        services = []
        for space in spaces:
            space['apps'] = []
            temp_apps = self.cloudfoundry.get_apps_in_space(space['guid'])
            for temp_app in temp_apps['resources']:
                service_bindings = self.cloudfoundry.get_app_service_bindings(
                    temp_app['metadata']['guid'])
                data = dict(list(temp_app['metadata'].items()) +
                            list(temp_app['entity'].items()))
                app_services = self._get_app_services_guids(service_bindings)
                if app_services:
                    data['service_bindings'] = app_services
                space['apps'].append(data)
            services.extend(self._parse_services(
                self.cloudfoundry.get_spaces_summary(space['guid'])))
        return services

    def _get_spaces(self):
        spaces = []
        for org in self._cached_organizations:
            temp_spaces = self.cloudfoundry.get_organization_spaces(org)
            for temp_space in temp_spaces['resources']:
                spaces.append(dict(list(temp_space['metadata'].items()) +
                                   list(temp_space['entity'].items())))
        return spaces

    @ds_utils.update_state_on_changed(SERVICES)
    def _translate_services(self, obj):
        LOG.debug("services: %s", obj)
        row_data = CloudFoundryV2Driver.convert_objs(
            obj, self.services_translator)
        return row_data

    @ds_utils.update_state_on_changed(ORGANIZATIONS)
    def _translate_organizations(self, obj):
        LOG.debug("organziations: %s", obj)

        # convert_objs needs the data structured a specific way so we
        # do this here. Perhaps we can improve convert_objs later to be
        # more flexiable.
        results = [dict(list(o['metadata'].items()) +
                        list(o['entity'].items()))
                   for o in obj['resources']]
        row_data = CloudFoundryV2Driver.convert_objs(
            results,
            self.organizations_translator)
        return row_data

    @ds_utils.update_state_on_changed(SPACES)
    def _translate_spaces(self, obj):
        LOG.debug("spaces: %s", obj)
        row_data = CloudFoundryV2Driver.convert_objs(
            obj,
            self.spaces_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.cloudfoundry, action, action_args)



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
cfg.CONF.distributed_architecture = True

from congress import exception as congressException
from congress.tests import base
from congress.tests import fake_datasource
from congress.tests import helper


class TestDataSource(base.SqlTestCase):

    def setUp(self):
        super(TestDataSource, self).setUp()
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])
        self.dseNode = helper.make_dsenode_new_partition('testnode')

    def _get_datasource_request(self):
        # leave ID out--generated during creation
        return {'name': 'aaron',
                'driver': 'fake_datasource',
                'description': 'hello world!',
                'enabled': True,
                'type': None,
                'config': {'auth_url': 'foo',
                           'username': 'armax',
                           'password': 'password',
                           'tenant_name': 'armax'}}

    def test_add_datasource(self):
        req = self._get_datasource_request()
        result = self.dseNode.add_datasource(req)
        # test equality of return value except for 'id' field
        del(result['id'])
        self.assertEqual(req, result)
        # check that service actually on dseNode
        services = self.dseNode.get_services()
        self.assertEqual(len(services), 1)
        self.assertEqual(services[0].service_id, req['name'])
        self.assertTrue(isinstance(services[0],
                        fake_datasource.FakeDataSource))
        obj = self.dseNode.invoke_service_rpc(
            req['name'], 'get_status', source_id=None, params=None)
        self.assertIsNotNone(obj)

    def test_get_datasource(self):
        req = self._get_datasource_request()
        ds = self.dseNode.add_datasource(req)
        result = self.dseNode.get_datasource(ds['id'])
        # test equality except for 'id' field
        del(result['id'])
        self.assertEqual(req, result)

    def test_get_datasources(self):
        req = self._get_datasource_request()
        self.dseNode.add_datasource(req)
        result = self.dseNode.get_datasources()
        self.assertEqual(len(result), 1)
        result = result[0]
        # test equality except for 'id' field
        del(result['id'])
        self.assertEqual(req, result)

    def test_get_datasources2(self):
        req1 = self._get_datasource_request()
        req1['name'] = 'datasource1'
        result1 = self.dseNode.add_datasource(req1)
        req2 = self._get_datasource_request()
        req2['name'] = 'datasource2'
        result2 = self.dseNode.add_datasource(req2)
        # check results of add_datasource
        for key, value in req1.items():
            self.assertEqual(value, result1[key])
        for key, value in req2.items():
            self.assertEqual(value, result2[key])
        # check services actually on dseNode
        services = self.dseNode.get_services()
        self.assertEqual(len(services), 2)
        self.assertEqual(set([s.service_id for s in services]),
                         set(['datasource1', 'datasource2']))
        self.assertTrue(isinstance(services[0],
                        fake_datasource.FakeDataSource))
        self.assertTrue(isinstance(services[1],
                        fake_datasource.FakeDataSource))
        # check results of get_datasources
        resultall = self.dseNode.get_datasources()
        self.assertEqual(len(resultall), 2)
        # check equality except for 'id' field
        byname = {x['name']: x for x in resultall}
        for x in byname.values():
            del(x['id'])
        self.assertEqual(byname, {'datasource1': req1, 'datasource2': req2})

    def test_get_datasources_hide_secret(self):
        req = self._get_datasource_request()
        self.dseNode.add_datasource(req)
        result = self.dseNode.get_datasources(filter_secret=True)
        result = result[0]
        # check equality except that 'config'/'password' is hidden
        req['config']['password'] = "<hidden>"
        del(result['id'])
        self.assertEqual(result, req)

    def test_create_datasource_duplicate_name(self):
        req = self._get_datasource_request()
        self.dseNode.add_datasource(req)
        self.assertRaises(congressException.DatasourceNameInUse,
                          self.dseNode.add_datasource, req)

    def test_delete_datasource(self):
        req = self._get_datasource_request()
        result = self.dseNode.add_datasource(req)
        self.dseNode.delete_datasource(result)
        # check that service is actually deleted
        services = self.dseNode.get_services()
        self.assertEqual(len(services), 0)
        self.assertRaises(
            congressException.NotFound, self.dseNode.invoke_service_rpc,
            req['name'], 'get_status', source_id=None, params=None)
        # TODO(thinrichs): test that we've actually removed
        #   the row from the DB

    # TODO(dse2): this test relies on coordination between dseNode and
    #  policy engine.  Much harder in distributed system.  Need to decide
    #  if we want that kind of invariant and if so implement it.
    # def test_delete_datasource_error(self):
    #     req = self._get_datasource_request()
    #     req['driver'] = 'fake_datasource'
    #     req['config'] = {'auth_url': 'foo',
    #                      'username': 'armax',
    #                      'password': 'password',
    #                      'tenant_name': 'armax'}
    #     # let driver generate this for us.
    #     del req['id']
    #     result = self.datasource_mgr.add_datasource(req)
    #     engine = self.dseNode.service_object('engine')
    #     engine.create_policy('alice')
    #     engine.insert('p(x) :- %s:q(x)' % req['name'], 'alice')
    #     self.assertRaises(exception.DanglingReference,
    #                       self.datasource_mgr.delete_datasource,
    #                       result['id'])

    def test_delete_invalid_datasource(self):
        req = self._get_datasource_request()
        req['id'] = 'fake-id'
        self.assertRaises(congressException.DatasourceNotFound,
                          self.dseNode.delete_datasource, req)

    # TODO(dse2): Doesn't seem like we need this (or it will be moved to API).
    # def test_get_driver_schema(self):
    #     schema = self.datasource_mgr.get_driver_schema(
    #         'fake_datasource')
    #     self.assertEqual(
    #         schema,
    #         fake_datasource.FakeDataSource.get_schema())

    def test_duplicate_driver_name_raises(self):
        # Load the driver twice
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource',
             'congress.tests.fake_datasource.FakeDataSource'])
        self.assertRaises(congressException.BadConfig,
                          self.dseNode.load_drivers)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import socket
import subprocess
import tempfile

from oslo_log import log as logging
from tempest.common import credentials_factory as credentials
from tempest import config
from tempest.lib import exceptions
from tempest import manager as tempestmanager
from tempest import test

from congress_tempest_tests.services.policy import policy_client
from congress_tempest_tests.tests.scenario import helper
from congress_tempest_tests.tests.scenario import manager_congress

CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestHA(manager_congress.ScenarioPolicyBase):

    def setUp(self):
        super(TestHA, self).setUp()
        self.keypairs = {}
        self.servers = []
        self.replicas = {}
        self.services_client = self.admin_manager.identity_services_client
        self.endpoints_client = self.admin_manager.endpoints_client

    def _prepare_replica(self, port_num):
        replica_url = "http://127.0.0.1:%d" % port_num
        resp = self.services_client.create_service(
            'congressha',
            CONF.congressha.replica_type,
            description='policy ha service')
        self.replica_service_id = resp['OS-KSADM:service']['id']
        resp = self.endpoints_client.create_endpoint(
            self.replica_service_id,
            CONF.identity.region,
            publicurl=replica_url,
            adminurl=replica_url,
            internalurl=replica_url)
        self.replica_endpoint_id = resp['endpoint']['id']

    def _cleanup_replica(self):
        self.endpoints_client.delete_endpoint(self.replica_endpoint_id)
        self.services_client.delete_service(self.replica_service_id)

    def start_replica(self, port_num):
        self._prepare_replica(port_num)
        f = tempfile.NamedTemporaryFile(mode='w', suffix='.conf',
                                        prefix='congress%d-' % port_num,
                                        dir='/tmp', delete=False)
        conf_file = f.name
        template = open('/etc/congress/congress.conf')
        conf = template.read()

        # Add 'bind_port' and 'datasource_sync_period' to conf file.
        index = conf.find('[DEFAULT]') + len('[DEFAULT]\n')
        conf = (conf[:index] + 'bind_port = %d\n' % port_num +
                'datasource_sync_period = 5\n' + conf[index:])
        sindex = conf.find('signing_dir')
        conf = conf[:sindex] + '#' + conf[sindex:]

        f.write(conf)
        f.close()

        args = ['/usr/bin/python',
                'bin/congress-server',
                '--config-file',
                conf_file]
        out = tempfile.NamedTemporaryFile(mode='w', suffix='.out',
                                          prefix='congress%d-' % port_num,
                                          dir='/tmp', delete=False)
        err = tempfile.NamedTemporaryFile(mode='w', suffix='.err',
                                          prefix='congress%d-' % port_num,
                                          dir='/tmp', delete=False)
        p = subprocess.Popen(args, stdout=out, stderr=err,
                             cwd=helper.root_path())

        assert port_num not in self.replicas
        self.replicas[port_num] = (p, conf_file)

    def stop_replica(self, port_num):
        proc, conf_file = self.replicas[port_num]
        # Using proc.terminate() will block at proc.wait(), no idea why yet
        proc.kill()
        proc.wait()
        os.unlink(conf_file)
        self.replicas[port_num] = (None, conf_file)
        self._cleanup_replica()

    def create_client(self, client_type):
        creds = credentials.get_configured_credentials('identity_admin')
        auth_prov = tempestmanager.get_auth_provider(creds)

        return policy_client.PolicyClient(
            auth_prov, client_type,
            CONF.identity.region)

    def datasource_exists(self, client, datasource_id):
        try:
            LOG.debug("datasource_exists begin")
            body = client.list_datasource_status(datasource_id)
            LOG.debug("list_datasource_status: %s", str(body))
        except exceptions.NotFound as e:
            LOG.debug("not found")
            return False
        except exceptions.Unauthorized as e:
            LOG.debug("connection refused")
            return False
        except socket.error as e:
            LOG.debug("Replica server not ready")
            return False
        except Exception as e:
            raise e
        return True

    def datasource_missing(self, client, datasource_id):
        try:
            LOG.debug("datasource_missing begin")
            body = client.list_datasource_status(datasource_id)
            LOG.debug("list_datasource_status: %s", str(body))
        except exceptions.NotFound as e:
            LOG.debug("not found")
            return True
        except exceptions.Unauthorized as e:
            LOG.debug("connection refused")
            return False
        except socket.error as e:
            LOG.debug("Replica server not ready")
            return False
        except Exception as e:
            raise e
        return False

    def find_fake(self, client):
        datasources = client.list_datasources()
        for r in datasources['results']:
            if r['name'] == 'fake':
                LOG.debug('existing fake driver: %s', str(r['id']))
                return r['id']
        return None

    def create_fake(self, client):
        # Create fake datasource if it does not exist.  Returns the
        # fake datasource id.
        fake_id = self.find_fake(client)
        if fake_id:
            return fake_id

        item = {'id': None,
                'name': 'fake',
                'driver': 'fake_datasource',
                'config': '{"username":"fakeu", "tenant_name": "faket",' +
                          '"password": "fakep",' +
                          '"auth_url": "http://127.0.0.1:5000/v2"}',
                'description': 'bar',
                'enabled': True}
        ret = client.create_datasource(item)
        LOG.debug('created fake driver: %s', str(ret['id']))
        return ret['id']

    @test.attr(type='smoke')
    def test_datasource_db_sync_add(self):
        # Verify that a replica adds a datasource when a datasource
        # appears in the database.
        client1 = self.admin_manager.congress_client

        # delete fake if it exists.
        old_fake_id = self.find_fake(client1)
        if old_fake_id:
            client1.delete_datasource(old_fake_id)

        # Verify that primary server has no fake datasource
        if not test.call_until_true(
                func=lambda: self.datasource_missing(client1, old_fake_id),
                duration=60, sleep_for=1):
            raise exceptions.TimeoutException(
                "primary should not have fake, but does")

        need_to_delete_fake = False
        try:
            # Create a new fake datasource
            fake_id = self.create_fake(client1)
            need_to_delete_fake = True

            # Verify that primary server has fake datasource
            if not test.call_until_true(
                    func=lambda: self.datasource_exists(client1, fake_id),
                    duration=60, sleep_for=1):
                raise exceptions.TimeoutException(
                    "primary should have fake, but does not")

            # start replica
            self.start_replica(CONF.congressha.replica_port)

            # Create session for second server.
            client2 = self.create_client(CONF.congressha.replica_type)

            # Verify that second server has fake datasource
            if not test.call_until_true(
                    func=lambda: self.datasource_exists(client2, fake_id),
                    duration=60, sleep_for=1):
                raise exceptions.TimeoutException(
                    "replica should have fake, but does not")

            # Remove fake from primary server instance.
            LOG.debug("removing fake datasource %s", str(fake_id))
            client1.delete_datasource(fake_id)
            need_to_delete_fake = False

            # Confirm that fake is gone from primary server instance.
            if not test.call_until_true(
                    func=lambda: self.datasource_missing(client1, fake_id),
                    duration=60, sleep_for=1):
                self.stop_replica(CONF.congressha.replica_port)
                raise exceptions.TimeoutException(
                    "primary instance still has fake")
            LOG.debug("removed fake datasource from primary instance")

            # Confirm that second service instance removes fake.
            if not test.call_until_true(
                    func=lambda: self.datasource_missing(client2, fake_id),
                    duration=60, sleep_for=1):
                raise exceptions.TimeoutException(
                    "replica should remove fake, but still has it")

        finally:
            self.stop_replica(CONF.congressha.replica_port)
            if need_to_delete_fake:
                self.admin_manager.congress_client.delete_datasource(fake_id)

    @test.attr(type='smoke')
    def test_datasource_db_sync_remove(self):
        # Verify that a replica removes a datasource when a datasource
        # disappears from the database.
        client1 = self.admin_manager.congress_client
        fake_id = self.create_fake(client1)
        need_to_delete_fake = True
        try:
            self.start_replica(CONF.congressha.replica_port)

            # Verify that primary server has fake datasource
            if not test.call_until_true(
                    func=lambda: self.datasource_exists(client1, fake_id),
                    duration=60, sleep_for=1):
                raise exceptions.TimeoutException(
                    "primary should have fake, but does not")

            # Create session for second server.
            client2 = self.create_client(CONF.congressha.replica_type)

            # Verify that second server has fake datasource
            if not test.call_until_true(
                    func=lambda: self.datasource_exists(client2, fake_id),
                    duration=60, sleep_for=1):
                raise exceptions.TimeoutException(
                    "replica should have fake, but does not")

            # Remove fake from primary server instance.
            LOG.debug("removing fake datasource %s", str(fake_id))
            client1.delete_datasource(fake_id)
            need_to_delete_fake = False

            # Confirm that fake is gone from primary server instance.
            if not test.call_until_true(
                    func=lambda: self.datasource_missing(client1, fake_id),
                    duration=60, sleep_for=1):
                self.stop_replica(CONF.congressha.replica_port)
                raise exceptions.TimeoutException(
                    "primary instance still has fake")
            LOG.debug("removed fake datasource from primary instance")

            # Confirm that second service instance removes fake.
            if not test.call_until_true(
                    func=lambda: self.datasource_missing(client2, fake_id),
                    duration=60, sleep_for=1):
                raise exceptions.TimeoutException(
                    "replica should remove fake, but still has it")

        finally:
            self.stop_replica(CONF.congressha.replica_port)
            if need_to_delete_fake:
                self.admin_manager.congress_client.delete_datasource(fake_id)



# Copyright (c) 2015 Huawei, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

from six.moves import http_client
import webob

from congress.tests import base
from congress.tests import fake_wsgi


class TestVersions(base.TestCase):

    def setUp(self):
        super(TestVersions, self).setUp()

    def test_versions_list(self):
        req = webob.Request.blank('/')
        req.accept = "application/json"
        res = req.get_response(fake_wsgi.wsgi_app())
        self.assertEqual(http_client.OK, res.status_int)
        self.assertEqual("application/json", res.content_type)
        versions = json.loads(res.body.decode('utf-8'))
        expected = {
            "versions": [{
                "status": "CURRENT",
                "updated": "2013-08-12T17:42:13Z",
                "id": "v1",
                "links": [{
                    "href": "http://localhost/v1/",
                    "rel": "self"
                }]
            }]
        }
        self.assertEqual(expected, versions)

    def test_versions_choices(self):
        req = webob.Request.blank('/fake')
        req.accept = "application/json"
        res = req.get_response(fake_wsgi.wsgi_app())
        self.assertEqual(http_client.MULTIPLE_CHOICES, res.status_int)
        self.assertEqual("application/json", res.content_type)
        versions = json.loads(res.body.decode('utf-8'))
        expected = {
            "choices": [{
                "status": "CURRENT",
                "updated": "2013-08-12T17:42:13Z",
                "id": "v1",
                "links": [{
                    "href": "http://localhost/v1/fake",
                    "rel": "self"
                }]
            }]
        }
        self.assertEqual(expected, versions)

    def test_version_v1_show(self):
        req = webob.Request.blank('/v1')
        req.accept = "application/json"
        res = req.get_response(fake_wsgi.wsgi_app())
        self.assertEqual(http_client.OK, res.status_int)
        self.assertEqual("application/json", res.content_type)
        versions = json.loads(res.body.decode('utf-8'))
        expected = {
            "version": {
                "status": "CURRENT",
                "updated": "2013-08-12T17:42:13Z",
                "id": "v1",
                "links": [{
                    "href": "http://localhost/v1/",
                    "rel": "self"
                }, {
                    "rel": "describedby",
                    "type": "text/html",
                    "href": "http://congress.readthedocs.org/",
                }]
            }
        }
        self.assertEqual(expected, versions)

    def test_version_v1_multiple_path(self):
        req = webob.Request.blank('/v1')
        req.accept = "application/json"
        res = req.get_response(fake_wsgi.wsgi_app())
        self.assertEqual(http_client.OK, res.status_int)
        self.assertEqual("application/json", res.content_type)

        req_ = webob.Request.blank('/v1/')
        req_.accept = "application/json"
        res_ = req_.get_response(fake_wsgi.wsgi_app())
        self.assertEqual(http_client.OK, res_.status_int)
        self.assertEqual("application/json", res_.content_type)

        self.assertEqual(json.loads(res.body.decode('utf-8')),
                         json.loads(res_.body.decode('utf-8')))

    def test_version_v1_not_found(self):
        req = webob.Request.blank('/v1/fake')
        req.accept = "application/json"
        res = req.get_response(fake_wsgi.wsgi_app())
        self.assertEqual(http_client.NOT_FOUND, res.status_int)
        self.assertEqual("application/json", res.content_type)



# Copyright 2015 Symantec.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Congress Policy Middleware.

"""
import json

from nova.i18n import _
from nova import wsgi
from oslo_config import cfg
from oslo_log import log as logging
import webob.dec
import webob.exc

# policy enforcement flow
from congressclient.v1 import client
import keystoneclient
from keystoneclient.v3 import client as ksv3client
from novaclient.v2 import client as nova_v2

LOG = logging.getLogger(__name__)


class Congress(wsgi.Middleware):
    """Make a request context from keystone headers."""

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):

        if req.environ['REQUEST_METHOD'] != 'POST':
            return self.application

        raw_path = req.environ['RAW_PATH_INFO']

        if "metadata" in raw_path:
            return self.application

        if "servers/action" in raw_path:
            return self.application

        flavor_ref = json.loads(req.body)['server']['flavorRef']

        token = req.environ['HTTP_X_AUTH_TOKEN']

        tenant_name = req.environ['HTTP_X_TENANT_NAME']

        CONF = cfg.CONF

        # obtain identity endpoint url
        url = CONF.keystone_authtoken.auth_url

        # obtain one of support keystone api versions
        raw_versions = keystoneclient.discover.available_versions(url,
                                                                  session=None)
        version = raw_versions[-1]['id']

        # assemble auth_url
        auth_url = url + '/' + version

        auth = keystoneclient.auth.identity.v2.Token(
            auth_url=auth_url,
            token=token, tenant_name=tenant_name)

        session = keystoneclient.session.Session(auth=auth)
        congress = client.Client(session=session,
                                 auth=None,
                                 interface='publicURL',
                                 service_type='policy')

        # Aggregating resource usage within domain level
        domain = req.environ['HTTP_X_PROJECT_DOMAIN_NAME']

        # obtain list of projects under this domain
        k3_client = ksv3client.Client(session=session)

        projects = k3_client.projects.list(domain=domain)
        # obtain list of hosts under each of these projects

        nova_c = nova_v2.Client(session=session)
        ram_p = 0
        disk_p = 0
        cpus_p = 0
        for project in projects:

            search_opts = {
                'all_tenants': 1,
                'tenant_id': project.id,
            }

            servers_p = nova_c.servers.list(search_opts=search_opts)

            # locate flavor of each host
            for server in servers_p:

                info = nova_c.servers.get(server=server)
                flavor_id = info._info['flavor']['id']
                fd = nova_c.flavors.get(flavor=flavor_id)
                ram_p += fd.ram
                disk_p += fd.disk
                disk_p += fd.ephemeral
                cpus_p += fd.vcpus

        # incrementally add each type of resource
        # assemble query policy based on the data-usage
        # with memory_p, disk_p and cpus_p

        fd = nova_c.flavors.get(flavor=flavor_ref)
        ram_p += fd.ram
        disk_p += fd.disk
        disk_p += fd.ephemeral
        cpus_p += fd.vcpus
        domain_resource = ("(" + domain + "," + str(ram_p) + "," +
                           str(disk_p) + "," + str(cpus_p) + ")")

        validation_result = congress.execute_policy_action(
            "classification",
            "simulate",
            False,
            True,
            {'query': 'domain_resource_usage_exceeded (domain)',
             # this needs to be defined in congress server
             'action_policy': 'nova_quota_action',
             'sequence': 'domain_resource+'+domain_resource})

        if validation_result["result"]:

            messages = validation_result["result"]

            if messages:
                result_str = "\n  ".join(map(str, messages))
                msg = _(
                    "quota is not sufficient for this VM deployment").format(
                    "\n  " + result_str)
                LOG.error(msg)

            LOG.debug(messages)
            return webob.exc.HTTPUnauthorized(explanation=msg)
        else:
            LOG.info('Model valid')

        return self.application



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import eventlet
import sys

from oslo_config import cfg
from oslo_messaging import conffixture

from congress.dse2.data_service import DataService
from congress.dse2.dse_node import DseNode
from congress.tests import base


# For manual testing, support using rabbit driver instead of fake
USE_RABBIT = False
if len(sys.argv) > 1:
    driver_flg = sys.argv[1].lower()
    if driver_flg == '--rabbit':
        USE_RABBIT = True
    elif driver_flg != '--fake':
        print("Usage: %s [--fake | --rabbit]" % sys.argv[0])
        sys.exit(1)
    sys.argv[1:] = sys.argv[2:]


class TestControlBus(base.TestCase):

    def setUp(self):
        super(TestControlBus, self).setUp()

        if USE_RABBIT:
            self.messaging_config = cfg.CONF
        else:
            mc_fixture = conffixture.ConfFixture(cfg.CONF)
            mc_fixture.conf.transport_url = 'kombu+memory://'
            self.messaging_config = mc_fixture.conf
        self.messaging_config.rpc_response_timeout = 1

    def tearDown(self):
        super(TestControlBus, self).tearDown()

    def test_control_bus_discovery(self):
        nodes = []
        services = []

        def _create_node_with_services(num):
            nid = 'cbd_node%s' % num
            nodes.append(DseNode(self.messaging_config, nid, []))
            ns = []
            for s in range(num):
                # intentionally starting different number services
                ns.append(DataService('cbd-%d_svc-%d' % (num, s)))
                nodes[-1].register_service(ns[-1])
            services.append(ns)
            return nodes[-1]

        for i in range(3):
            n = _create_node_with_services(i)
            n.start()

        eventlet.sleep(.1)  # Allow for heartbeat propagation

        def _validate_peer_view(node):
            status = node.dse_status()
            expected_peers = set([n.node_id for n in nodes
                                  if n.node_id != node.node_id])
            peers = set(status['peers'].keys())
            self.assertEqual(peers, expected_peers,
                             '%s has incorrect peers list' % node.node_id)
            for n in nodes:
                if n.node_id == node.node_id:
                    continue
                expected_services = [s.service_id for s in n._services]
                services = [s['service_id']
                            for s in status['peers'][n.node_id]['services']]
                self.assertEqual(set(expected_services), set(services),
                                 '%s has incorrect service list'
                                 % node.node_id)

        for n in nodes:
            _validate_peer_view(n)

        # Late arriving node
        n = _create_node_with_services(3)
        n.start()
        eventlet.sleep(.1)  # Allow for heartbeat propagation
        for n in nodes:
            _validate_peer_view(n)


# TODO(pballand): replace with congress unit test framework when convenient
if __name__ == '__main__':
    import unittest
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_congress
----------------------------------

Tests for `congress` module.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from functools import reduce

import os

import mock
from mox3 import mox
import neutronclient.v2_0
from oslo_log import log as logging

from congress.api import webservice
from congress.common import config
from congress.datalog import base as datalog_base
from congress.datalog import compile
from congress import harness
from congress.tests import base
import congress.tests.datasources.test_neutron_driver as test_neutron
from congress.tests import helper


LOG = logging.getLogger(__name__)


class TestCongress(base.SqlTestCase):

    def setUp(self):
        """Setup tests that use multiple mock neutron instances."""
        super(TestCongress, self).setUp()
        # create neutron mock and tell cage to use that mock
        #  https://code.google.com/p/pymox/wiki/MoxDocumentation
        mock_factory = mox.Mox()
        neutron_mock = mock_factory.CreateMock(
            neutronclient.v2_0.client.Client)
        neutron_mock2 = mock_factory.CreateMock(
            neutronclient.v2_0.client.Client)

        config_override = {'neutron2': {'username': 'demo', 'tenant_name':
                                        'demo', 'password': 'password',
                                        'auth_url':
                                            'http://127.0.0.1:5000/v2.0',
                                        'module':
                                            'datasources/neutron_driver.py'},
                           'nova': {'username': 'demo',
                                    'tenant_name': 'demo',
                                    'password': 'password',
                                    'auth_url': 'http://127.0.0.1:5000/v2.0',
                                    'module': 'datasources/nova_driver.py'},
                           'neutron': {'username': 'demo',
                                       'tenant_name': 'demo',
                                       'password': 'password',
                                       'auth_url':
                                            'http://127.0.0.1:5000/v2.0',
                                       'module':
                                            'datasources/neutron_driver.py'}}

        cage = harness.create(helper.root_path(), config_override)
        # Disable synchronizer because the this test creates
        # datasources without also inserting them into the database.
        # The synchronizer would delete these datasources.
        cage.service_object('synchronizer').set_poll_time(0)

        engine = cage.service_object('engine')

        api = {'policy': cage.service_object('api-policy'),
               'rule': cage.service_object('api-rule'),
               'table': cage.service_object('api-table'),
               'row': cage.service_object('api-row'),
               'datasource': cage.service_object('api-datasource'),
               'status': cage.service_object('api-status'),
               'schema': cage.service_object('api-schema')}

        config = {'username': 'demo',
                  'auth_url': 'http://127.0.0.1:5000/v2.0',
                  'tenant_name': 'demo',
                  'password': 'password',
                  'module': 'datasources/neutron_driver.py',
                  'poll_time': 0}

        # FIXME(arosen): remove all this code
        # monkey patch
        engine.create_policy('neutron',
                             kind=datalog_base.DATASOURCE_POLICY_TYPE)
        engine.create_policy('neutron2',
                             kind=datalog_base.DATASOURCE_POLICY_TYPE)
        engine.create_policy('nova',
                             kind=datalog_base.DATASOURCE_POLICY_TYPE)
        harness.load_data_service(
            'neutron', config, cage,
            os.path.join(helper.root_path(), "congress"), 1)
        service = cage.service_object('neutron')
        engine.set_schema('neutron', service.get_schema())
        harness.load_data_service(
            'neutron2', config, cage,
            os.path.join(helper.root_path(), "congress"), 2)

        engine.set_schema('neutron2', service.get_schema())
        config['module'] = 'datasources/nova_driver.py'
        harness.load_data_service(
            'nova', config, cage,
            os.path.join(helper.root_path(), "congress"), 3)
        engine.set_schema('nova', service.get_schema())

        cage.service_object('neutron').neutron = neutron_mock
        cage.service_object('neutron2').neutron = neutron_mock2
        # delete all policies that aren't builtin, so we have clean slate
        names = set(engine.policy_names()) - engine.builtin_policy_names
        for name in names:
            try:
                api['policy'].delete_item(name, {})
            except KeyError:
                pass

        # Turn off schema checking
        engine.module_schema = None

        # initialize neutron_mocks
        network1 = test_neutron.network_response
        port_response = test_neutron.port_response
        router_response = test_neutron.router_response
        sg_group_response = test_neutron.security_group_response
        neutron_mock.list_networks().InAnyOrder().AndReturn(network1)
        neutron_mock.list_ports().InAnyOrder().AndReturn(port_response)
        neutron_mock.list_routers().InAnyOrder().AndReturn(router_response)
        neutron_mock.list_security_groups().InAnyOrder().AndReturn(
            sg_group_response)
        neutron_mock2.list_networks().InAnyOrder().AndReturn(network1)
        neutron_mock2.list_ports().InAnyOrder().AndReturn(port_response)
        neutron_mock2.list_routers().InAnyOrder().AndReturn(router_response)
        neutron_mock2.list_security_groups().InAnyOrder().AndReturn(
            sg_group_response)
        mock_factory.ReplayAll()

        self.cage = cage
        self.engine = engine
        self.api = api

    def setup_config(self):
        args = ['--config-file', helper.etcdir('congress.conf.test')]
        config.init(args)

    def test_startup(self):
        """Test that everything is properly loaded at startup."""
        engine = self.engine
        api = self.api
        helper.retry_check_subscriptions(
            engine, [(api['rule'].name, 'policy-update')])
        helper.retry_check_subscribers(
            api['rule'], [(engine.name, 'policy-update')])

    def test_synchronize_policy_no_erratic_change(self):
        """Test that synchronize_policies does not changes init state"""
        with mock.patch.object(self.engine, 'delete_policy') as d:
            with mock.patch.object(self.engine, 'create_policy') as c:
                self.engine.synchronizer.synchronize_policies()
                # TODO(ekcs): How can we show args used if erratic call made?
                d.assert_not_called()
                c.assert_not_called()

    def test_policy_subscriptions(self):
        """Test that policy engine subscriptions adjust to policy changes."""
        engine = self.engine
        api = self.api
        cage = self.cage
        policy = engine.DEFAULT_THEORY

        # Send formula
        formula = test_neutron.create_network_group('p')
        LOG.debug("Sending formula: %s", formula)
        api['rule'].publish(
            'policy-update', [compile.Event(formula, target=policy)])
        # check we have the proper subscriptions
        self.assertTrue('neutron' in cage.services)
        neutron = cage.service_object('neutron')
        helper.retry_check_subscriptions(engine, [('neutron', 'networks')])
        helper.retry_check_subscribers(neutron, [(engine.name, 'networks')])

    def test_neutron(self):
        """Test polling and publishing of neutron updates."""
        engine = self.engine
        api = self.api
        cage = self.cage
        policy = engine.DEFAULT_THEORY

        # Send formula
        formula = test_neutron.create_network_group('p')
        LOG.debug("Sending formula: %s", formula)
        api['rule'].publish(
            'policy-update', [compile.Event(formula, target=policy)])
        helper.retry_check_nonempty_last_policy_change(engine)
        LOG.debug("All services: %s", cage.services.keys())
        neutron = cage.service_object('neutron')
        neutron.poll()
        ans = ('p("240ff9df-df35-43ae-9df5-27fae87f2492") ')
        helper.retry_check_db_equal(engine, 'p(x)', ans, target=policy)

    def test_multiple(self):
        """Test polling and publishing of multiple neutron instances."""
        api = self.api
        cage = self.cage
        engine = self.engine
        policy = engine.DEFAULT_THEORY

        # Send formula
        formula = test_neutron.create_networkXnetwork_group('p')
        api['rule'].publish(
            'policy-update', [compile.Event(formula, target=policy)])
        helper.retry_check_nonempty_last_policy_change(engine)
        # poll datasources
        neutron = cage.service_object('neutron')
        neutron2 = cage.service_object('neutron2')
        neutron.poll()
        neutron2.poll()
        # check answer
        ans = ('p("240ff9df-df35-43ae-9df5-27fae87f2492",  '
               '  "240ff9df-df35-43ae-9df5-27fae87f2492") ')
        helper.retry_check_db_equal(engine, 'p(x,y)', ans, target=policy)

    def test_datasource_api_model(self):
        """Test the datasource api model.

        Same as test_multiple except we use the api interface
        instead of the DSE interface.
        """
        self.skipTest("Move to test/api/api_model and use fake driver...")
        # FIXME(arosen): we should break out these tests into
        # congress/tests/api/test_datasource.py
        with mock.patch("congress.managers.datasource.DataSourceDriverManager."
                        "get_datasource_drivers_info") as get_info:
            get_info.return_value = [{'datasource_driver': 'neutron'},
                                     {'datasource_driver': 'neutron2'},
                                     {'datasource_driver': 'nova'}]
            api = self.api
            engine = self.engine
            # Insert formula (which creates neutron services)
            net_formula = test_neutron.create_networkXnetwork_group('p')
            LOG.debug("Sending formula: %s", net_formula)
            context = {'policy_id': engine.DEFAULT_THEORY}
            api['rule'].add_item(
                {'rule': str(net_formula)}, {}, context=context)
            datasources = api['datasource'].get_items({})['results']
            datasources = [d['datasource_driver'] for d in datasources]
            self.assertEqual(set(datasources),
                             set(['neutron', 'neutron2', 'nova']))

    def test_row_api_model(self):
        """Test the row api model."""
        self.skipTest("Move to test/api/test_row_api_model..")
        api = self.api
        engine = self.engine
        # add some rules defining tables
        context = {'policy_id': engine.DEFAULT_THEORY}
        api['rule'].add_item(
            {'rule': 'p(x) :- q(x)'},
            {}, context=context)
        api['rule'].add_item(
            {'rule': 'p(x) :- r(x)'},
            {}, context=context)
        api['rule'].add_item(
            {'rule': 'q(x) :- r(x)'},
            {}, context=context)
        api['rule'].add_item(
            {'rule': 'r(1) :- true'},
            {}, context=context)

        # without tracing
        context['table_id'] = 'p'
        ans = api['row'].get_items({}, context=context)
        s = frozenset([tuple(x['data']) for x in ans['results']])
        t = frozenset([(1,)])
        self.assertEqual(s, t, "Rows without tracing")
        self.assertTrue('trace' not in ans, "Rows should have no Trace")
        self.assertEqual(len(ans['results']), 1)  # no duplicates

        # with tracing
        ans = api['row'].get_items({'trace': 'true'}, context=context)
        s = frozenset([tuple(x['data']) for x in ans['results']])
        t = frozenset([(1,)])
        self.assertEqual(s, t, "Rows with tracing")
        self.assertTrue('trace' in ans, "Rows should have trace")
        self.assertEqual(len(ans['trace'].split('\n')), 16)

        # unknown policy table
        context = {'policy_id': engine.DEFAULT_THEORY, 'table_id': 'unktable'}
        ans = api['row'].get_items({}, context=context)
        self.assertEqual(len(ans['results']), 0)

        # unknown policy
        context = {'policy_id': 'unkpolicy', 'table_id': 'unktable'}
        ans = api['row'].get_items({}, context=context)
        self.assertEqual(len(ans['results']), 0)

        # unknown datasource table
        context = {'ds_id': 'neutron', 'table_id': 'unktable'}
        ans = api['row'].get_items({}, context=context)
        self.assertEqual(len(ans['results']), 0)

        # unknown datasource
        context = {'ds_id': 'unkds', 'table_id': 'unktable'}
        ans = api['row'].get_items({}, context=context)
        self.assertEqual(len(ans['results']), 0)

    def test_policy_api_model_execute(self):
        def _execute_api(client, action, action_args):
            LOG.info("_execute_api called on %s and %s", action, action_args)
            positional_args = action_args['positional']
            named_args = action_args['named']
            method = reduce(getattr, action.split('.'), client)
            method(*positional_args, **named_args)

        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def _get_testkey(self):
                return self.testkey

            def disconnectNetwork(self, arg1, arg2, arg3):
                self.testkey = "arg1=%s arg2=%s arg3=%s" % (arg1, arg2, arg3)

        nova_client = NovaClient("testing")
        nova = self.cage.service_object('nova')
        nova._execute_api = _execute_api
        nova.nova_client = nova_client

        api = self.api
        body = {'name': 'nova:disconnectNetwork',
                'args': {'positional': ['value1', 'value2'],
                         'named': {'arg3': 'value3'}}}

        request = helper.FakeRequest(body)
        result = api['policy'].execute_action({}, {}, request)
        self.assertEqual(result, {})

        expected_result = "arg1=value1 arg2=value2 arg3=value3"
        f = nova.nova_client._get_testkey
        helper.retry_check_function_return_value(f, expected_result)

    def test_rule_insert_delete(self):
        self.api['policy'].add_item({'name': 'alice'}, {})
        context = {'policy_id': 'alice'}
        (id1, _) = self.api['rule'].add_item(
            {'rule': 'p(x) :- plus(y, 1, x), q(y)'}, {}, context=context)
        ds = self.api['rule'].get_items({}, context)['results']
        self.assertEqual(len(ds), 1)
        self.api['rule'].delete_item(id1, {}, context)
        ds = self.engine.policy_object('alice').content()
        self.assertEqual(len(ds), 0)

    # TODO(thinrichs): Clean up this file.  In particular, make it possible
    #   to group all of the policy-execute tests into their own class.
    # Execute[...] tests
    def test_policy_execute(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self, arg1):
                LOG.info("disconnectNetwork called on %s", arg1)
                self.testkey = "arg1=%s" % arg1

        nova_client = NovaClient(None)
        nova = self.cage.service_object('nova')
        nova.nova_client = nova_client

        # insert rule and data
        self.api['policy'].add_item({'name': 'alice'}, {})
        (id1, _) = self.api['rule'].add_item(
            {'rule': 'execute[nova:disconnectNetwork(x)] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        (id2, _) = self.api['rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # insert more data
        self.api['rule'].add_item(
            {'rule': 'q(2)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)
        ans = "arg1=2"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # insert irrelevant data
        self.api['rule'].add_item(
            {'rule': 'r(3)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

        # delete relevant data
        self.api['rule'].delete_item(
            id2, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

        # delete policy rule
        self.api['rule'].delete_item(
            id1, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

    def test_policy_execute_data_first(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self, arg1):
                LOG.info("disconnectNetwork called on %s", arg1)
                self.testkey = "arg1=%s" % arg1

        nova_client = NovaClient(None)
        nova = self.cage.service_object('nova')
        nova.nova_client = nova_client

        # insert rule and data
        self.api['policy'].add_item({'name': 'alice'}, {})
        self.api['rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        self.api['rule'].add_item(
            {'rule': 'execute[nova:disconnectNetwork(x)] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

    def test_policy_execute_dotted(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey
                self.servers = ServersClass()

        class ServersClass(object):
            def __init__(self):
                self.ServerManager = ServerManagerClass()

        class ServerManagerClass(object):
            def __init__(self):
                self.testkey = None

            def pause(self, id_):
                self.testkey = "arg1=%s" % id_

        nova_client = NovaClient(None)
        nova = self.cage.service_object('nova')
        nova.nova_client = nova_client

        self.api['policy'].add_item({'name': 'alice'}, {})
        self.api['rule'].add_item(
            {'rule': 'execute[nova:servers.ServerManager.pause(x)] :- q(x)'},
            {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        self.api['rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: nova.nova_client.servers.ServerManager.testkey
        helper.retry_check_function_return_value(f, ans)

    def test_policy_execute_no_args(self):
        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self):
                LOG.info("disconnectNetwork called")
                self.testkey = "noargs"

        nova_client = NovaClient(None)
        nova = self.cage.service_object('nova')
        nova.nova_client = nova_client

        # Note: this probably isn't the behavior we really want.
        #  But at least we have a test documenting that behavior.

        # insert rule and data
        self.api['policy'].add_item({'name': 'alice'}, {})
        (id1, rule1) = self.api['rule'].add_item(
            {'rule': 'execute[nova:disconnectNetwork()] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        (id2, rule2) = self.api['rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "noargs"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # insert more data (which DOES NOT cause an execution)
        (id3, rule3) = self.api['rule'].add_item(
            {'rule': 'q(2)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)

        # delete all data
        self.api['rule'].delete_item(
            id2, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)

        self.api['rule'].delete_item(
            id3, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)

        # insert data (which now DOES cause an execution)
        (id4, rule3) = self.api['rule'].add_item(
            {'rule': 'q(3)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)
        ans = "noargs"
        f = lambda: nova.nova_client.testkey
        helper.retry_check_function_return_value(f, ans)

        # delete policy rule
        self.api['rule'].delete_item(
            id1, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 2)

    def test_datasource_request_refresh(self):
        # Remember that neutron does not poll automatically here, which
        #   is why this test actually testing request_refresh
        neutron = self.cage.service_object('neutron')
        LOG.info("neutron.state: %s", neutron.state)
        self.assertEqual(len(neutron.state['ports']), 0)
        # TODO(thinrichs): Seems we can't test the datasource API at all.
        # api['datasource'].request_refresh_action(
        #     {}, context, helper.FakeRequest({}))
        neutron.request_refresh()
        f = lambda: len(neutron.state['ports'])
        helper.retry_check_function_return_value_not_eq(f, 0)

    def test_neutron_policy_execute(self):
        class NeutronClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def disconnectNetwork(self, arg1):
                LOG.info("disconnectNetwork called on %s", arg1)
                self.testkey = "arg1=%s" % arg1

        neutron_client = NeutronClient(None)
        neutron = self.cage.service_object('neutron')
        neutron.neutron = neutron_client

        # insert rule and data
        self.api['policy'].add_item({'name': 'alice'}, {})
        (id1, _) = self.api['rule'].add_item(
            {'rule': 'execute[neutron:disconnectNetwork(x)] :- q(x)'}, {},
            context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 0)
        (id2, _) = self.api['rule'].add_item(
            {'rule': 'q(1)'}, {}, context={'policy_id': 'alice'})
        self.assertEqual(len(self.engine.logger.messages), 1)
        ans = "arg1=1"
        f = lambda: neutron.neutron.testkey
        helper.retry_check_function_return_value(f, ans)

    def test_datasource_api_model_execute(self):
        def _execute_api(client, action, action_args):
            positional_args = action_args.get('positional', [])
            named_args = action_args.get('named', {})
            method = reduce(getattr, action.split('.'), client)
            method(*positional_args, **named_args)

        class NovaClient(object):
            def __init__(self, testkey):
                self.testkey = testkey

            def _get_testkey(self):
                return self.testkey

            def disconnect(self, arg1, arg2, arg3):
                self.testkey = "arg1=%s arg2=%s arg3=%s" % (arg1, arg2, arg3)

            def disconnect_all(self):
                self.testkey = "action_has_no_args"

        nova_client = NovaClient("testing")
        nova = self.cage.service_object('nova')
        nova._execute_api = _execute_api
        nova.nova_client = nova_client

        execute_action = self.api['datasource'].execute_action

        # Positive test: valid body args, ds_id
        context = {'ds_id': 'nova'}
        body = {'name': 'disconnect',
                'args': {'positional': ['value1', 'value2'],
                         'named': {'arg3': 'value3'}}}
        request = helper.FakeRequest(body)
        result = execute_action({}, context, request)
        self.assertEqual(result, {})
        expected_result = "arg1=value1 arg2=value2 arg3=value3"
        f = nova.nova_client._get_testkey
        helper.retry_check_function_return_value(f, expected_result)

        # Positive test: no body args
        context = {'ds_id': 'nova'}
        body = {'name': 'disconnect_all'}
        request = helper.FakeRequest(body)
        result = execute_action({}, context, request)
        self.assertEqual(result, {})
        expected_result = "action_has_no_args"
        f = nova.nova_client._get_testkey
        helper.retry_check_function_return_value(f, expected_result)

        # Negative test: invalid ds_id
        context = {'ds_id': 'unknown_ds'}
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, request)

        # Negative test: no ds_id
        context = {}
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, request)

        # Negative test: empty body
        context = {'ds_id': 'nova'}
        bad_request = helper.FakeRequest({})
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, bad_request)

        # Negative test: no body name/action
        context = {'ds_id': 'nova'}
        body = {'args': {'positional': ['value1', 'value2'],
                         'named': {'arg3': 'value3'}}}
        bad_request = helper.FakeRequest(body)
        self.assertRaises(webservice.DataModelException, execute_action,
                          {}, context, bad_request)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
import novaclient

from congress.datalog import compile
from congress.datasources import nova_driver
from congress.dse import d6cage
from congress import exception
from congress.tests import base
from congress.tests.datasources import fakes
from congress.tests import helper


class TestNovaDriver(base.TestCase):

    def setUp(self):
        super(TestNovaDriver, self).setUp()
        nova_client = mock.MagicMock()
        self.nova = fakes.NovaFakeClient()
        with mock.patch.object(novaclient.client.Client, '__init__',
                               return_value=nova_client):
            self.driver = nova_driver.NovaDriver(
                name='nova',
                args=helper.datasource_openstack_args())

    def test_driver_called(self):
        self.assertIsNotNone(self.driver.nova_client)

    def test_servers(self):
        servers_raw = self.nova.servers.list(detailed=True)
        self.driver._translate_servers(servers_raw)
        server_tuples = self.driver.state[self.driver.SERVERS]

        self.assertEqual(3, len(server_tuples))
        #  tuple = (s.id, s.name, s.hostId, s.status, s.tenant_id,
        #   s.user_id, image, flavor)
        for t in server_tuples:
            id = t[0]
            name = t[1]
            host_id = t[2]
            status = t[3]
            tenant_id = t[4]
            user_id = t[5]
            image_id = t[6]
            flavor_id = t[7]
            zone = t[8]
            host_name = t[9]
            self.assertIn(id, [1234, 5678, 9012])
            # see congress.datasources.tests.unit.fakes for actual values
            if id == 1234:
                self.assertEqual("sample-server", name)
                self.assertEqual("e4d909c290d0fb1ca068ffaddf22cbd0", host_id)
                self.assertEqual("BUILD", status)
                self.assertEqual("33ea0494-2bdf-4382-a445-9068997430b9",
                                 user_id)
                self.assertEqual("50e14867-7c64-4ec9-be8d-ed2470ca1d24",
                                 tenant_id)
                self.assertEqual(2, image_id)
                self.assertEqual(1, flavor_id)
                self.assertEqual('default', zone)
                self.assertEqual('host1', host_name)

            elif id == 5678:
                self.assertEqual("sample-server2", name)
                self.assertEqual("9e107d9d372bb6826bd81d3542a419d6", host_id)
                self.assertEqual("ACTIVE", status)
                self.assertEqual("33ea0494-2bdf-4382-a445-9068997430b9",
                                 user_id)
                self.assertEqual("50e14867-7c64-4ec9-be8d-ed2470ca1d24",
                                 tenant_id)
                self.assertEqual(2, image_id)
                self.assertEqual(1, flavor_id)
                self.assertEqual('None', zone)
                self.assertEqual('None', host_name)

            elif id == 9012:
                self.assertEqual("sample-server3", name)
                self.assertEqual("9e107d9d372bb6826bd81d3542a419d6", host_id)
                self.assertEqual("ACTIVE", status)
                self.assertEqual("33ea0494-2bdf-4382-a445-9068997430b9",
                                 user_id)
                self.assertEqual("50e14867-7c64-4ec9-be8d-ed2470ca1d24",
                                 tenant_id)
                self.assertEqual(2, image_id)
                self.assertEqual(1, flavor_id)
                self.assertEqual('foo', zone)
                self.assertEqual('host2', host_name)

    def test_flavors(self):
        flavor_raw = self.nova.flavors.list(detailed=True)
        self.driver._translate_flavors(flavor_raw)

        flavor_tuples = self.driver.state[self.driver.FLAVORS]

        self.assertEqual(4, len(flavor_tuples))
        # "id", "name", "vcpus", "ram", "disk", "ephemeral",
        #            "rxtx_factor")
        for f in flavor_tuples:
            id = f[0]
            name = f[1]
            vcpus = f[2]
            ram = f[3]
            disk = f[4]
            ephemeral = f[5]
            rxtx_factor = f[6]

            self.assertIn(id, [1, 2, 3, 4])

            # {'id': 1, 'name': '256 MB Server', 'ram': 256, 'disk': 10,
            # 'vcpus' : 1, 'OS-FLV-EXT-DATA:ephemeral': 10,
            # 'os-flavor-access:is_public': True, 'rxtx_factor' : 1.0,
            # 'links': {}},
            if id == 1:
                self.assertEqual('256 MB Server', name)
                self.assertEqual(256, ram)
                self.assertEqual(10, disk)
                self.assertEqual(1, vcpus)
                self.assertEqual(10, ephemeral)
                self.assertEqual(1.0, rxtx_factor)
            # {'id': 2, 'name': '512 MB Server', 'ram': 512, 'disk': 20,
            #  'vcpus' :2, 'OS-FLV-EXT-DATA:ephemeral': 20,
            #  'os-flavor-access:is_public': False, 'rxtx_factor' : 1.0,
            #  'links': {}},
            elif id == 2:
                self.assertEqual('512 MB Server', name)
                self.assertEqual(512, ram)
                self.assertEqual(20, disk)
                self.assertEqual(2, vcpus)
                self.assertEqual(20, ephemeral)
                self.assertEqual(1.0, rxtx_factor)
            # {'id': 3, 'name': '128 MB Server', 'ram': 128, 'disk': 0,
            #  'vcpus' : 4, 'OS-FLV-EXT-DATA:ephemeral': 0,
            #  'os-flavor-access:is_public': True, 'rxtx_factor' : 3.0,
            #  'links': {}}
            elif id == 3:
                self.assertEqual('128 MB Server', name)
                self.assertEqual(128, ram)
                self.assertEqual(0, disk)
                self.assertEqual(4, vcpus)
                self.assertEqual(0, ephemeral)
                self.assertEqual(3.0, rxtx_factor)
            # {'id': 4, 'name': '1024 MB Server', 'ram': 1024, 'disk': 10,
            #  'vcpus' : 3, 'OS-FLV-EXT-DATA:ephemeral': 10,
            #  'os-flavor-access:is_public': True, 'rxtx_factor' : 2.0,
            #  'links': {}},
            elif id == 4:
                self.assertEqual('1024 MB Server', name)
                self.assertEqual(1024, ram)
                self.assertEqual(10, disk)
                self.assertEqual(3, vcpus)
                self.assertEqual(10, ephemeral)
                self.assertEqual(2.0, rxtx_factor)
                self.assertEqual('1024 MB Server', name)

    def test_hosts(self):
        host_list = self.nova.hosts.list()
        self.driver._translate_hosts(host_list)
        host_tuples = self.driver.state[self.driver.HOSTS]
        self.assertEqual(2, len(host_tuples))
        # {'hosts':
        #      [{'host_name': 'host1',
        #        'service': 'nova-compute',
        #        'zone': zone},
        #       {'host_name': 'host2',
        #        'service': 'nova-cert',
        #        'zone': zone}]}
        for host in host_tuples:
            host_name = host[0]
            service = host[1]
            zone = host[2]

            if host_name == 'host1':
                self.assertEqual('nova-compute', service)
                self.assertEqual('nova1', str(zone))
            elif host_name == 'host2':
                self.assertEqual('nova-cert', service)
                self.assertEqual('nova1', str(zone))

    def test_services(self):
        service_list = self.nova.services.list()
        self.driver._translate_services(service_list)
        expected_ret = {
            1: [1, 'nova-compute', 'nova', 'nova1', 'enabled', 'up',
                '2015-07-28T08:28:37.000000', 'None'],
            2: [2, 'nova-schedule', 'nova', 'nova1', 'disabled', 'up',
                '2015-07-28T08:28:38.000000', 'daily maintenance']
        }
        service_tuples = self.driver.state[self.driver.SERVICES]

        self.assertEqual(2, len(service_tuples))

        for s in service_tuples:
            map(self.assertEqual, expected_ret[s[0]], s)

    def test_availability_zones(self):
        az_list = self.nova.availability_zones.list()
        self.driver._translate_availability_zones(az_list)
        expected_ret = {
            'AZ1': ['AZ1', 'available'],
            'AZ2': ['AZ2', 'not available']
        }
        az_tuples = self.driver.state[self.driver.AVAILABILITY_ZONES]

        self.assertEqual(2, len(az_tuples))

        for az in az_tuples:
            map(self.assertEqual, expected_ret[az[0]], az)

    def test_communication(self):
        """Test for communication.

        Test the module's ability to be loaded into the DSE
        by checking its ability to communicate on the message bus.
        """
        cage = d6cage.d6Cage()

        # Create modules.
        # Turn off polling so we don't need to deal with real data.
        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        cage.loadModule("NovaDriver",
                        helper.data_module_path("nova_driver.py"))
        cage.loadModule("PolicyDriver", helper.policy_module_path())
        cage.createservice(name="policy", moduleName="PolicyDriver",
                           args={'d6cage': cage,
                                 'rootdir': helper.data_module_path(''),
                                 'log_actions_only': True})
        cage.createservice(name="nova", moduleName="NovaDriver", args=args)

        # Check that data gets sent from nova to policy as expected
        nova = cage.service_object('nova')
        policy = cage.service_object('policy')
        policy.debug_mode()
        policy.create_policy('nova')
        policy.set_schema('nova', compile.Schema({'server': (1,)}))
        policy.subscribe('nova', 'server',
                         callback=policy.receive_data)

        # publishing is slightly convoluted b/c deltas are computed
        #  automatically.  (Not just convenient--useful so that DSE
        #  properly handles the initial state problem.)
        # Need to set nova.state and nova.prior_state and then publish
        #  anything.

        # publish server(1), server(2), server(3)
        helper.retry_check_subscribers(nova, [(policy.name, 'server')])
        nova.prior_state = {}
        nova.state['server'] = set([(1,), (2,), (3,)])
        nova.publish('server', None)
        helper.retry_check_db_equal(
            policy, 'nova:server(x)',
            'nova:server(1) nova:server(2) nova:server(3)')

        # publish server(1), server(4), server(5)
        nova.prior_state['server'] = nova.state['server']
        nova.state['server'] = set([(1,), (4,), (5,)])
        nova.publish('server', None)
        helper.retry_check_db_equal(
            policy, 'nova:server(x)',
            'nova:server(1) nova:server(4) nova:server(5)')

    # TODO(thinrichs): test that Nova's polling functionality
    #   works properly.  Or perhaps could bundle this into the
    #   tests above if we check self.state results.
    #   See Neutron's test_polling

    def test_execute(self):
        class NovaClient(object):
            def __init__(self):
                self.testkey = None

            def connectNetwork(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        nova_client = NovaClient()
        self.driver.nova_client = nova_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('connectNetwork', api_args)

        self.assertEqual(expected_ans, nova_client.testkey)

    def test_execute_servers_set_meta(self):
        class server(object):
            def __init__(self):
                self.testkey = None

            def set_meta(self, server=None, metadata=None):
                self.testkey = 'server=%s, metadata=%s' % (server, metadata)

        class NovaClient(object):
            def __init__(self):
                self.servers = server()

        nova_client = NovaClient()
        self.driver.nova_client = nova_client
        expected_ans = "server=1, metadata={'meta-key1': 'meta-value1'}"

        action_args = {'positional': ['1', 'meta-key1', 'meta-value1']}
        self.driver.execute('servers_set_meta', action_args)

        self.assertEqual(expected_ans, nova_client.servers.testkey)

    def test_execute_with_non_executable_method(self):
        action_args = {'positional': ['1', 'meta-key1', 'meta-value1']}
        self.assertRaises(exception.CongressException,
                          self.driver.execute,
                          'get_nova_credentials_v2', action_args)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from six.moves import range

from congress.datasources import datasource_driver


def d6service(name, keys, inbox, datapath, args):
    """Create a dataservice instance.

    This method is called by d6cage to create a dataservice
    instance.  There are a couple of parameters we found useful
    to add to that call, so we included them here instead of
    modifying d6cage (and all the d6cage.createservice calls).
    """
    return BenchmarkDriver(name, keys, inbox, datapath, args)


class BenchmarkDriver(datasource_driver.PollingDataSourceDriver):
    BENCHTABLE = 'benchtable'
    value_trans = {'translation-type': 'VALUE'}
    translator = {
        'translation-type': 'HDICT',
        'table-name': BENCHTABLE,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'field1', 'translator': value_trans},
             {'fieldname': 'field2', 'translator': value_trans})}

    TRANSLATORS = [translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(BenchmarkDriver, self).__init__(name, keys,
                                              inbox, datapath, args)
        # used by update_from_datasources to manufacture data. Default small.
        self.datarows = 10
        self._init_end_start_poll()

    def update_from_datasource(self):
        self.state = {}
        # TODO(sh): using self.convert_objs() takes about 10x the time. Needs
        # optimization efforts.
        row_data = tuple((self.BENCHTABLE, ('val1_%d' % i, 'val2_%d' % i))
                         for i in range(self.datarows))
        for table, row in row_data:
            if table not in self.state:
                self.state[table] = set()
            self.state[table].add(row)

    def get_credentials(self, *args, **kwargs):
        return {}



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add name attribute to policy rules

Revision ID: 56e86d51ec62
Revises: 532e9e1f0f3a
Create Date: 2015-01-14 13:08:53.945019

"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# revision identifiers, used by Alembic.
revision = '56e86d51ec62'
down_revision = '532e9e1f0f3a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('policy_rules',
                  sa.Column('name', sa.String(length=255), nullable=True))


def downgrade():
    op.drop_column('policy_rules', 'name')



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import subprocess
import time

from oslo_log import log as logging
import pulp

from congress.datalog import arithmetic_solvers
from congress.datalog import base
from congress.datalog import compile
from congress.datalog import nonrecursive
from congress import exception
from congress.policy_engines import base_driver

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return ComputePlacementEngine(name, keys, inbox, datapath, args)


# TODO(thinrichs): Figure out what to move to the base class PolicyEngineDriver
#  Could also pull out the Datalog-to-LP conversion, potentially.
class ComputePlacementEngine(base_driver.PolicyEngineDriver):
    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(ComputePlacementEngine, self).__init__(
            name, keys, inbox, datapath)
        self.policy = nonrecursive.MultiModuleNonrecursiveRuleTheory(name=name)
        self.initialized = True
        self.guest_host_assignment = {}
        self.lplang = arithmetic_solvers.PulpLpLang()
        self.vm_migrator = VmMigrator()

    ###########################
    # Policy engine interface

    def insert(self, formula):
        return self.policy.insert(self.parse1(formula))

    def delete(self, formula):
        return self.policy.delete(self.parse1(formula))

    def select(self, query):
        ans = self.policy.select(self.parse1(query))
        return " ".join(str(x) for x in ans)

    def set_policy(self, policy):
        LOG.info("%s:: setting policy to %s", str(self.name), str(policy))
        # empty out current policy
        external = [compile.Tablename.build_service_table(service, name)
                    for service, name in self._current_external_tables()]
        self.policy.empty(tablenames=external, invert=True)

        # insert new policy and subscribe to the tablenames referencing a
        #    datasource driver
        for rule in self.parse(policy):
            self.policy.insert(rule)
        LOG.info("new policy: %s", self.policy.content_string())

        # initialize table subscriptions
        self.initialize_table_subscriptions()

        # enforce policy
        self.enforce_policy()

    def initialize_table_subscriptions(self):
        """Initialize table subscription.

        Once policies have all been loaded, this function subscribes to
        all the necessary tables.  See UPDATE_TABLE_SUBSCRIPTIONS as well.
        """
        tablenames = self.policy.tablenames()
        tablenames = [compile.Tablename.parse_service_table(table)
                      for table in tablenames]
        tablenames = [(service, name) for (service, name) in tablenames
                      if service is not None]
        self._set_subscriptions(tablenames)

    def _set_subscriptions(self, tablenames):
        """Update subscriptions on DSE to be exactly @tablenames."""
        subscriptions = set(self._current_external_tables())
        tablenames = set(tablenames)
        toadd = tablenames - subscriptions
        torem = subscriptions - tablenames
        for service, tablename in toadd:
            if service is not None:
                LOG.info("%s:: subscribing to (%s, %s)",
                         self.name, service, tablename)
                self.subscribe(service, tablename,
                               callback=self.receive_data)

        for service, tablename in torem:
            if service is not None:
                LOG.info("%s:: unsubscribing from (%s, %s)",
                         self.name, service, tablename)
                self.unsubscribe(service, tablename)
                relevant_tables = [compile.Tablename.build_service_table(
                                   service, tablename)]
                self.policy.empty(relevant_tables)

    def _current_external_tables(self):
        """Return list of tables engine is currently subscribed to."""
        return [(value.key, value.dataindex)
                for value in self.subdata.values()]

    ################################################################
    # Receiving data published on the DSE by other services
    # For PoC, assuming all data already present and no pubs.
    #   So we're ignoring this for now.

    def receive_data(self, msg):
        """Event handler for when a dataservice publishes data.

        That data can either be the full table (as a list of tuples)
        or a delta (a list of Events).
        """
        LOG.info("%s:: received data msg %s", self.name, msg)
        # if empty data, assume it is an init msg, since noop otherwise
        if len(msg.body.data) == 0:
            self.receive_data_full(msg)
        else:
            # grab an item from any iterable
            dataelem = next(iter(msg.body.data))
            if isinstance(dataelem, compile.Event):
                self.receive_data_update(msg)
            else:
                self.receive_data_full(msg)
        self.enforce_policy()

    def receive_data_full(self, msg):
        """Handler for when dataservice publishes full table."""
        LOG.info("%s:: received full data msg for %s: %s",
                 self.name, msg.header['dataindex'],
                 ";".join(str(x) for x in msg.body.data))
        tablename = compile.Tablename.build_service_table(
            msg.replyTo, msg.header['dataindex'])

        # Use a generator to avoid instantiating all these Facts at once.
        #   Don't print out 'literals' since that will eat the generator
        literals = (compile.Fact(tablename, row) for row in msg.body.data)

        LOG.info("%s:: begin initialize_tables %s", self.name, tablename)
        self.policy.initialize_tables([tablename], literals)
        LOG.info("%s:: end initialize data msg for %s", self.name, tablename)
        select = [str(x) for x in self.select('p(x)')]
        LOG.info("%s:: select('p(x)'): %s ENDED", self.name, " ".join(select))

    def receive_data_update(self, msg):
        """Handler for when dataservice publishes a delta."""
        LOG.info("%s:: received update data msg for %s: %s",
                 self.name, msg.header['dataindex'],
                 ";".join(str(x) for x in msg.body.data))
        new_events = []
        for event in msg.body.data:
            assert compile.is_atom(event.formula), (
                "receive_data_update received non-atom: " +
                str(event.formula))
            # prefix tablename with data source
            actual_table = compile.Tablename.build_service_table(
                msg.replyTo, event.formula.table.table)
            values = [term.name for term in event.formula.arguments]
            newevent = compile.Event(compile.Fact(actual_table, values),
                                     insert=event.insert)
            new_events.append(newevent)
        (permitted, changes) = self.policy.update(new_events)
        if not permitted:
            raise exception.CongressException(
                "Update not permitted." + '\n'.join(str(x) for x in changes))
        else:
            tablename = msg.header['dataindex']
            service = msg.replyTo
            LOG.debug("update data msg for %s from %s caused %d "
                      "changes: %s", tablename, service, len(changes),
                      ";".join(str(x) for x in changes))

    #######################################
    # Policy enforcement

    def enforce_policy(self):
        """Enforce policy by migrating VMs to minimize warnings.

        Raises LpProblemUnsolvable if the LP cannot solve the
        given problem.

        Raises LpConversionFailure if self.policy cannot be converted
        into an LP problem.
        """
        LOG.info("Enforcing policy")
        ans = self.policy.select(self.parse1('warning(x)'), True)
        if len(ans) == 0:
            return
        # grab assignment
        g_h_assignment = self.calculate_vm_assignment()
        self.guest_host_assignment = dict(g_h_assignment)
        # migrate
        for guest in g_h_assignment:
            g_h_assignment[guest] = [g_h_assignment[guest], 0]
        self.vm_migrator.do_migrations(g_h_assignment)

    def calculate_vm_assignment(self):
        """Calculate where VMs should be located in order to minimize warnings.

        Returns a dictionary from guest ID to host ID where that guest should
        be located.

        Raises LpProblemUnsolvable if the LP cannot solve the
        given problem.

        Raises LpConversionFailure if self.policy cannot be converted
        into an LP problem.
        """

        g_h_assignment = {}
        LOG.info("* Calculating VM assignment for Datalog policy: *")
        LOG.info(self.policy.content_string())
        migproblem, value_mapping = self.policy_to_lp_problem()
        LOG.info("* Converted to PuLP program: *")
        LOG.info("problem: %s", migproblem)
        migproblem.solve()
        LOG.info("problem status: %s", migproblem.status)
        if pulp.LpStatus[migproblem.status] == 'Optimal':
            LOG.info("value-mapping: %s", value_mapping)
            for var in migproblem.variables():
                LOG.info("var: %s = %s", var.name, var.varValue)
                if var.name.startswith('assign'):
                    g, h = var.name.lstrip('assign').lstrip('_').split('_')
                    g = value_mapping.get(int(g), g)
                    h = value_mapping.get(int(h), h)
                    LOG.info("guest %s, host %s has value %s",
                             g, h, var.varValue)
                    if var.varValue == 1.0:
                        # add correct old host
                        g_h_assignment[g] = h

            return g_h_assignment
        raise LpProblemUnsolvable(str(migproblem))

    #######################################
    # Toplevel conversion of Datalog to LP

    # mapping Datalog tables to LP decision variables

    def policy_to_lp_problem(self):
        """Return an LP problem representing the state of this engine.

        Returns an instance of self.lplang.problem representing the policy
        and the current data of this engine.
        """
        opt, hard = self.policy_to_lp()
        LOG.info("* Converted Datalog policy to DatalogLP *")
        LOG.info("optimization:\n%s", opt)
        LOG.info("constraints:\n%s", "\n".join(str(x) for x in hard))
        bounds = {}
        for exp in hard:
            self.set_bounds(exp, bounds)
        return self.lplang.problem(opt, hard, bounds)

    def policy_to_lp(self):
        """Transform self.policy into a (non-)linear programming problem.

        Returns (<optimization criteria>, <hard constraints>) where
        each are represented using expressions constructed by self.lplang.
        """
        # soft constraints. optimization criteria: minimize number of warnings
        # LOG.info("* Converting warning(x) to DatalogLP *")
        wquery = self.parse1('warning(x)')
        warnings, wvars = self.datalog_to_lp(wquery, [])
        opt = self.lplang.makeOr(*wvars)
        # hard constraints.  all must be false
        # LOG.info("* Converting error(x) to DatalogLP *")
        equery = self.parse1('error(x)')
        errors, evars = self.datalog_to_lp(equery, [])
        hard = [self.lplang.makeNotEqual(var, 1) for var in evars]
        # domain-specific axioms, e.g. sum of guest memory util = host mem util
        # LOG.info("* Constructing domain-specific axioms *")
        axioms = self.domain_axioms()
        return opt, warnings + errors + hard + axioms

    def set_bounds(self, expr, bounds):
        """Find upper bounds on all variables occurring in expr.

        :param expr is a LpLang.Expression
        :param bounds is a dictionary mapping an Expression's tuple() to a
            number.

        Modifies bounds to include values for all variables occurring inside
        expr.
        """
        # LOG.info("set_bounds(%s)", expr)
        variables = self.lplang.variables(expr)
        for var in variables:
            tup = var.tuple()
            if tup not in bounds:
                bounds[tup] = 10

    ##########################
    # Domain-specific axioms

    def domain_axioms(self):
        """Return a list of all the domain-specific axioms as strings.

        Axioms define relationships between LP decision variables that we
        would not expect the user to write.
        """
        # TODO(thinrichs): just defining relationship between mem-usage for
        #   guests and hosts.  Add rest of axioms.
        hosts = self.get_hosts()
        guests = self.get_guests()
        memusage = self.get_memusage()

        memusage_ax = self._domain_axiom_memusage(hosts, guests, memusage)
        assign_ax = self._domain_axiom_assignment(hosts, guests)
        return memusage_ax + assign_ax

    def _domain_axiom_assignment(self, hosts, guests):
        """Return axioms for assignment variables.

        :param hosts is the list of host IDs
        :param guests is the list of guest IDs

        assign[h1,g] + ... + assign[hn, g] = 1
        """
        axioms = []
        for g in guests:
            hostvars = [self._construct_assign(h, g) for h in hosts]
            axioms.append(self.lplang.makeEqual(
                1, self.lplang.makeArith('plus', *hostvars)))
        return axioms

    def _construct_assign(self, host, guest):
        return self.lplang.makeBoolVariable('assign', guest, host)

    def _domain_axiom_memusage(self, hosts, guests, memusage):
        """Return a list of LP axioms defining guest/host mem-usage.

        :param hosts is the list of host IDs
        :param guests is the list of guest IDs

        Axiom: sum of all guest mem-usage for those guests deployed on a host
        gives the mem-usage for that host:

        hMemUse[h] = assign[1][h]*gMemUse[1] + ... + assign[G][h]*gMemUse[G].

        Returns a list of LpLang expressions.
        Raises NotEnoughData if it does not have guest memory usage.
        """
        axioms = []

        for h in hosts:
            guest_terms = []
            for guest in guests:
                if guest not in memusage:
                    raise NotEnoughData(
                        "could not find guest mem usage: %s" % guest)
                guest_terms.append(
                    self.lplang.makeArith(
                        'times',
                        self._construct_assign(h, guest),
                        memusage[guest]))
            axioms.append(
                self.lplang.makeEqual(
                    self.lplang.makeIntVariable('hMemUse', h),
                    self.lplang.makeArith('plus', *guest_terms)))
        return axioms

    def get_hosts(self):
        query = self.parse1('nova:host(id, zone, memory_capacity)')
        host_rows = self.policy.select(query)
        return set([lit.arguments[0].name for lit in host_rows])

    def get_guests(self):
        query = self.parse1('nova:server(id, name, host)')
        guest_rows = self.policy.select(query)
        return set([lit.arguments[0].name for lit in guest_rows])

    def get_memusage(self):
        query = self.parse1('ceilometer:mem_consumption(id, mem)')
        rows = self.policy.select(query)
        return {lit.arguments[0].name: lit.arguments[1].name
                for lit in rows}

    #########################
    # Convert datalog to LP

    unknowns = ['ceilometer:mem_consumption']
    rewrites = ['ceilometer:mem_consumption(x, y) :- '
                'var("hMemUse", x), output(y)']

    def datalog_to_lp(self, query, unknown_table_possibilities):
        """Convert rules defining QUERY in self.policy into a linear program.

        @unknowns is the list of tablenames that should become
        decision variables.  @unknown_table_possibilities is the list
        of all possible instances of the decision variable tables.
        """
        # TODO(thinrichs): figure out if/when negation is handled properly

        # a list of rules, each of which has an instance of QUERY in the head
        #   and whose bodies are drawn from unknowns.
        rules = self.policy.abduce(query, self.unknowns)
        # LOG.info("interpolates:\n%s", "\n".join(str(x) for x in rules))
        if len(unknown_table_possibilities):
            rules = self.policy.instances(query, unknown_table_possibilities)
            # LOG.info("instances:\n%s", "\n".join(str(x) for x in rules))
        equalities, variables = self._to_lp(rules)
        # LOG.info("LP rules: \n%s", "\n".join(str(x) for x in equalities))
        # LOG.info("LP variables: %s", ", ".join(str(x) for x in variables))
        return equalities, variables

    def _to_lp(self, rules):
        """Compute an LP program equivalent to the given Datalog rules.

        :param rules: a list of Rule instances, all of which are ground
                      except for variables representing LP variables
        """
        # TODO(thinrichs): need type analysis to ensure we differentiate
        #    hosts from guests within ceilometer:mem_consumption
        act = nonrecursive.MultiModuleNonrecursiveRuleTheory()
        for var_rewrite_rule in self.rewrites:
            changes = act.insert(self.parse1(var_rewrite_rule))
            assert(changes)
        LOG.debug("action theory: %s", act.content_string())
        act.set_tracer(self.policy.tracer)
        definitions = {}
        for rule in rules:
            equalities, newrule = self._extract_lp_variable_equalities(
                rule, act)
            LOG.debug("equalities: %s", equalities)
            LOG.debug("newrule: %s", newrule)
            LOG.debug("newrule.body: %s", str(newrule.body))
            head = self._lit_to_lp_variable(newrule.head)
            LOG.debug("head: %s", str(head))
            LOG.debug("newrule.body: %s", newrule.body)
            body = []
            for lit in newrule.body:
                LOG.debug("processing %s", lit)
                body.append(self._lit_to_lp_arithmetic(lit, equalities))
            LOG.debug("new body: %s", ";".join(str(x) for x in body))
            conjunction = self.lplang.makeAnd(*body)
            LOG.debug("conjunct: %s", conjunction)
            if head not in definitions:
                definitions[head] = set([conjunction])
            else:
                definitions[head].add(conjunction)

        equalities = [self.lplang.makeEqual(h, self.lplang.makeOr(*bodies))
                      for h, bodies in definitions.items()]
        return equalities, definitions.keys()

    def _extract_lp_variable_equalities(self, rule, rewrite_theory):
        """Extract values for LP variables and slightly modify rule.

        :param rule: an instance of Rule
        :param rewrite_theory: reference to a theory that contains rules
               describing how tables correspond to LP variable inputs and
               outputs.

        Returns (i) dictionary mapping Datalog variable name (a string) to
        the set of LP variables to which it is equal and (ii) a rewriting
        of the rule that is the same as the original except some
        elements have been removed from the body.
        """
        newbody = []
        varnames = {}
        for lit in rule.body:
            result = self._extract_lp_variable_equality_lit(
                lit, rewrite_theory)
            if result is None:
                newbody.append(lit)
            else:
                datalogvar, lpvar = result
                if datalogvar not in varnames:
                    varnames[datalogvar] = set([lpvar])
                else:
                    varnames[datalogvar].add(lpvar)
        return varnames, compile.Rule(rule.head, newbody)

    def _extract_lp_variable_equality_lit(self, lit, rewrite_theory):
        """Identify datalog variable representing an LP-variable.

        :param lit: an instance of Literal
        :param rewrite_theory: reference to a theory that contains rules
               describing how tables correspond to LP variable inputs and
               outputs.
        Returns None, signifying literal does not include any datalog
        variable that maps to an LP variable, or (datalogvar, lpvar).
        """
        if lit.is_builtin():
            return
        # LOG.info("_extract_lp_var_eq_lit %s", lit)
        rewrites = rewrite_theory.abduce(lit, ['var', 'output'])
        # LOG.info("lit rewriting: %s", ";".join(str(x) for x in rewrites))
        if not rewrites:
            return
        assert(len(rewrites) == 1)
        varlit = next(lit for lit in rewrites[0].body
                      if lit.table.table == 'var')
        # LOG.info("varlit: %s", varlit)
        lpvar = self._varlit_to_lp_variable(varlit)
        outlit = next(lit for lit in rewrites[0].body
                      if lit.table.table == 'output')
        outvar = outlit.arguments[0].name
        # LOG.info("lpvar: %s; outvar: %s", lpvar, outvar)
        return outvar, lpvar

    def _lit_to_lp_arithmetic(self, lit, varnames):
        """Translates Datalog literal into an LP arithmetic statement.

        :param lit is a Literal instance and may include Datalog variables
        :param varnames is a dictionary from datalog variables to a set of
        LP variables

        Returns an LP arithmetic statement.

        Raises LpConversion if one of the Datalog variables appearing in
        lit has other than 1 value in varnames.
        Raises LpException if the arithmetic operator is not supported.
        """
        # TODO(thinrichs) translate to infix and use standard operators
        newargs = [self._term_to_lp_term(arg, varnames)
                   for arg in lit.arguments]
        return self.lplang.makeArith(lit.tablename(), *newargs)

    def _lit_to_lp_variable(self, lit):
        """Translates ground Datalog literal into an LP variable.

        :param lit is a Literal instance without variables
        Returns an LP variable.
        Raises LpConversionFailure if lit includes any Datalog variables.
        """
        if any(arg.is_variable() for arg in lit.arguments):
            raise self.lplang.LpConversionFailure(
                "Tried to convert literal %s into LP variable but "
                "found a Datalog variable" % lit)
        args = [arg.name for arg in lit.arguments]
        return self.lplang.makeVariable(lit.table.table, *args, type='bool')

    def _term_to_lp_term(self, term, varnames):
        """Translates Datalog term into an LP variable or a constant.

        :param term is an instance of Term
        :param varnames is a dictionary from varname to a set of LP variables

        Returns an LP variable, a number, or a string.

        Raises LpConversionFailure if Datalog variable appears without a
        corresponding LP variable or if multiple LP variables for a given
        Datalog variable.  (The latter condition could probably be handled
        without raising an error, but this is good for now.)
        """
        if term.is_variable():
            if term.name not in varnames:
                raise self.lplang.LpConversionFailure(
                    "Residual variable not assigned a value: %s" % term.name)
            if len(varnames[term.name]) > 1:
                raise self.lplang.LpConversionFailure(
                    "Variable name assigned to 2 different values: "
                    "%s assigned %s" % (term.name, varnames[term.name]))
            return next(iter(varnames[term.name]))
        return term.name

    def _varlit_to_lp_variable(self, lit):
        args = [x.name for x in lit.arguments[1:]]
        return self.lplang.makeVariable(lit.arguments[0].name, *args)

    #################
    # Miscellaneous

    def debug_mode(self):
        tracer = base.Tracer()
        tracer.trace('*')
        self.policy.set_tracer(tracer)

    def production_mode(self):
        tracer = base.Tracer()
        self.policy.set_tracer(tracer)

    def parse(self, policy):
        return compile.parse(policy, use_modules=False)

    def parse1(self, policy):
        return compile.parse1(policy, use_modules=False)


class NotEnoughData(exception.CongressException):
    pass


class LpProblemUnsolvable(exception.CongressException):
    pass


class VmMigrator(object):
    """Code for migrating VMs once we have a LP problem solution."""
    @classmethod
    def migrate(cls, guest, host):
        try:
            call = ["nova", "live-migration", str(guest), str(host)]
            LOG.info("migrating: %s", call)
            ret = subprocess.check_output(call, stderr=subprocess.STDOUT)
            if ret == 0:
                return True
        except Exception:
            pass

    @classmethod
    def check_status(cls, guest, host, status):
        g = subprocess.check_output(["nova", "list"])
        g = g.replace("-", "").replace("+", "").lstrip("[").rstrip("]")
        elems = g.split('\n')
        for elem in elems:
            e = elem.split("|")
            el = [x.strip() for x in e]
            try:
                if status == el[2]:
                    return True
            except Exception:
                pass

    @classmethod
    def do_migration(cls, guest, newh, oldh):
        if (newh == oldh):
            return True
        try:
            done = cls.migrate(guest, newh)
            if done:
                for i in range(3):
                    if cls.check_migrate(guest, newh, "ACTIVE"):
                        return True
                    else:
                        time.sleep(2)
        except Exception:
            pass
        return False

    # status: -1 if migration done
    @classmethod
    def getnext(cls, mapping, status):
        hi = max(status.values())
        if hi > 0:
            i = list(status.values()).index(hi)
            return list(status.keys())[i]

    @classmethod
    def do_migrations(cls, g_h_mapping):
        max_attempts = 10
        guest_mig_status = dict.fromkeys(g_h_mapping.keys(), max_attempts)
        g = cls.getnext(g_h_mapping, guest_mig_status)
        while g:
            newh, oldh = g_h_mapping[g]
            if cls.do_migration(g, newh, oldh):
                guest_mig_status[g] = -1
            else:
                guest_mig_status[g] -= 1
            g = cls.getnext(g_h_mapping, guest_mig_status)
        return guest_mig_status



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock

from congress.datasources import glancev2_driver
from congress.tests import base
from congress.tests import helper


class TestGlanceV2Driver(base.TestCase):

    def setUp(self):
        super(TestGlanceV2Driver, self).setUp()
        self.keystone_client_p = mock.patch(
            "keystoneclient.v2_0.client.Client")
        self.keystone_client_p.start()
        self.glance_client_p = mock.patch("glanceclient.v2.client.Client")
        self.glance_client_p.start()

        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        args['client'] = mock.MagicMock()
        self.driver = glancev2_driver.GlanceV2Driver(args=args)

        self.mock_images = {'images': [
            {u'checksum': u'9e486c3bf76219a6a37add392e425b36',
             u'container_format': u'bare',
             u'created_at': u'2014-10-01T20:28:08Z',
             u'disk_format': u'qcow2',
             u'file': u'/v2/images/c42736e7-8b09-4906-abd2-d6dc8673c297/file',
             u'id': u'c42736e7-8b09-4906-abd2-d6dc8673c297',
             u'min_disk': 0,
             u'min_ram': 0,
             u'name': u'Fedora-x86_64-20-20140618-sda',
             u'owner': u'4dfdcf14a20940799d89c7a5e7345978',
             u'protected': False,
             u'schema': u'/v2/schemas/image',
             u'size': 209649664,
             u'status': u'active',
             u'tags': ['type=xen2', 'type=xen'],
             u'updated_at': u'2014-10-01T20:28:09Z',
             u'visibility': u'public'},
            {u'checksum': u'4eada48c2843d2a262c814ddc92ecf2c',
             u'container_format': u'ami',
             u'created_at': u'2014-10-01T20:28:06Z',
             u'disk_format': u'ami',
             u'file': u'/v2/images/6934941f-3eef-43f7-9198-9b3c188e4aab/file',
             u'id': u'6934941f-3eef-43f7-9198-9b3c188e4aab',
             u'kernel_id': u'15ed89b8-588d-47ad-8ee0-207ed8010569',
             u'min_disk': 0,
             u'min_ram': 0,
             u'name': u'cirros-0.3.2-x86_64-uec',
             u'owner': u'4dfdcf14a20940799d89c7a5e7345978',
             u'protected': False,
             u'ramdisk_id': u'c244d5c7-1c83-414c-a90d-af7cea1dd3b5',
             u'schema': u'/v2/schemas/image',
             u'size': 25165824,
             u'status': u'active',
             u'tags': [],
             u'updated_at': u'2014-10-01T20:28:07Z',
             u'visibility': u'public'}]}

    def test_update_from_datasource(self):
        with mock.patch.object(self.driver.glance.images, "list") as img_list:
            img_list.return_value = self.mock_images['images']
            self.driver.update_from_datasource()
        expected = {'images': set([
            (u'6934941f-3eef-43f7-9198-9b3c188e4aab',
             u'active',
             u'cirros-0.3.2-x86_64-uec',
             u'ami',
             u'2014-10-01T20:28:06Z',
             u'2014-10-01T20:28:07Z',
             u'ami',
             u'4dfdcf14a20940799d89c7a5e7345978',
             'False',
             0,
             0,
             u'4eada48c2843d2a262c814ddc92ecf2c',
             25165824,
             u'/v2/images/6934941f-3eef-43f7-9198-9b3c188e4aab/file',
             u'15ed89b8-588d-47ad-8ee0-207ed8010569',
             u'c244d5c7-1c83-414c-a90d-af7cea1dd3b5',
             u'/v2/schemas/image',
             u'public'),
            (u'c42736e7-8b09-4906-abd2-d6dc8673c297',
             u'active',
             u'Fedora-x86_64-20-20140618-sda',
             u'bare',
             u'2014-10-01T20:28:08Z',
             u'2014-10-01T20:28:09Z',
             u'qcow2',
             u'4dfdcf14a20940799d89c7a5e7345978',
             'False',
             0,
             0,
             u'9e486c3bf76219a6a37add392e425b36',
             209649664,
             u'/v2/images/c42736e7-8b09-4906-abd2-d6dc8673c297/file',
             'None',
             'None',
             u'/v2/schemas/image',
             u'public')]),
            'tags': set([
                (u'c42736e7-8b09-4906-abd2-d6dc8673c297', 'type=xen'),
                (u'c42736e7-8b09-4906-abd2-d6dc8673c297', 'type=xen2')])}
        self.assertEqual(expected, self.driver.state)

    def test_execute(self):
        class GlanceClient(object):
            def __init__(self):
                self.testkey = None

            def createSnapshot(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        glance_client = GlanceClient()
        self.driver.glance = glance_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('createSnapshot', api_args)

        self.assertEqual(expected_ans, glance_client.testkey)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.api import base
from congress.api import error_codes
from congress.api import webservice
from congress import exception


LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    return RuleModel(name, keys, inbox=inbox, dataPath=datapath, **args)


class RuleModel(base.APIModel):
    """Model for handling API requests about policy Rules."""

    def policy_name(self, context):
        if 'ds_id' in context:
            return context['ds_id']
        elif 'policy_id' in context:
            return context['policy_id']

    def get_item(self, id_, params, context=None):
        """Retrieve item with id id_ from model.

        Args:
            id_: The ID of the item to retrieve
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The matching item or None if item with id_ does not exist.
        """
        try:
            args = {'id_': id_, 'policy_name': self.policy_name(context)}
            return self.invoke_rpc(self.engine, 'persistent_get_rule', args)
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)

    def get_items(self, params, context=None):
        """Get items in model.

        Args:
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """
        try:
            args = {'policy_name': self.policy_name(context)}
            rules = self.invoke_rpc(self.engine, 'persistent_get_rules', args)
            return {'results': rules}
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)

    def add_item(self, item, params, id_=None, context=None):
        """Add item to model.

        Args:
            item: The item to add to the model
            params: A dict-like object containing parameters
                    from the request query string and body.
            id_: The ID of the item, or None if an ID should be generated
            context: Key-values providing frame of reference of request

        Returns:
             Tuple of (ID, newly_created_item)

        Raises:
            KeyError: ID already exists.
        """
        if id_ is not None:
            raise webservice.DataModelException(
                *error_codes.get('add_item_id'))
        try:
            args = {'policy_name': self.policy_name(context),
                    'str_rule': item.get('rule'),
                    'rule_name': item.get('name'),
                    'comment': item.get('comment')}
            return self.invoke_rpc(self.engine, 'persistent_insert_rule', args)
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)

    def delete_item(self, id_, params, context=None):
        """Remove item from model.

        Args:
            id_: The ID of the item to be removed
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The removed item.

        Raises:
            KeyError: Item with specified id_ not present.
        """
        try:
            args = {'id_': id_, 'policy_name_or_id': self.policy_name(context)}
            return self.invoke_rpc(self.engine, 'persistent_delete_rule', args)
        except exception.CongressException as e:
            raise webservice.DataModelException.create(e)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock


class NovaFakeClient(mock.MagicMock):
    # TODO(rajdeepd): Replace Fake with mocks directly in test_neutron_driver
    def __init__(self, *args, **kwargs):
        super(NovaFakeClient, self).__init__(*args, **kwargs)
        self.servers = mock.MagicMock()
        self.servers.list.return_value = self.get_server_list()
        self.flavors = mock.MagicMock()
        self.flavors.list.return_value = self.get_flavor_list()

        self.hosts = mock.MagicMock()
        self.hosts.list.return_value = self.get_host_list()
        self.services = mock.MagicMock()
        self.services.list.return_value = self.get_service_list()

        self.availability_zones = mock.MagicMock()
        self.availability_zones.list.return_value = self.get_zone_list()

    def get_mock_server(self, id, name, host_id, status, tenant_id, user_id,
                        flavor, image, zone=None, host_name=None):
        server = mock.MagicMock()
        server.id = id
        server.hostId = host_id
        server.tenant_id = tenant_id
        server.user_id = user_id
        server.status = status
        server.name = name
        server.image = image
        server.flavor = flavor
        if zone is not None:
            setattr(server, 'OS-EXT-AZ:availability_zone', zone)
        else:
            # This ensures that the magic mock raises an AttributeError
            delattr(server, 'OS-EXT-AZ:availability_zone')
        if host_name is not None:
            setattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname',
                    host_name)
        else:
            # This ensures that the magic mock raises an AttributeError
            delattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname')
        return server

    def get_server_list(self):
        server_one = (
            self.get_mock_server(1234, 'sample-server',
                                 "e4d909c290d0fb1ca068ffaddf22cbd0",
                                 'BUILD',
                                 '50e14867-7c64-4ec9-be8d-ed2470ca1d24',
                                 '33ea0494-2bdf-4382-a445-9068997430b9',
                                 {"id": 1}, {"id": 2}, 'default', 'host1'))

        server_two = (
            self.get_mock_server(5678, 'sample-server2',
                                 "9e107d9d372bb6826bd81d3542a419d6",
                                 'ACTIVE',
                                 '50e14867-7c64-4ec9-be8d-ed2470ca1d24',
                                 '33ea0494-2bdf-4382-a445-9068997430b9',
                                 {"id": 1}, {"id": 2}))

        server_three = (
            self.get_mock_server(9012, 'sample-server3',
                                 "9e107d9d372bb6826bd81d3542a419d6",
                                 'ACTIVE',
                                 '50e14867-7c64-4ec9-be8d-ed2470ca1d24',
                                 '33ea0494-2bdf-4382-a445-9068997430b9',
                                 {"id": 1}, {"id": 2}, 'foo', 'host2'))

        return [server_one, server_two, server_three]

    def get_flavor(self, id, name, vcpus, ram, disk, ephemeral, rxtx_factor):
        f = mock.MagicMock()
        f.id = id
        f.name = name
        f.vcpus = vcpus
        f.ram = ram
        f.disk = disk
        f.ephemeral = ephemeral
        f.rxtx_factor = rxtx_factor
        return f

    def get_flavor_list(self):
        flavor_one = self.get_flavor(1, "256 MB Server", 1, 256, 10, 10, 1.0)
        flavor_two = self.get_flavor(2, "512 MB Server", 2, 512, 20, 20, 1.0)
        flavor_three = self.get_flavor(3, "128 MB Server", 4, 128, 0, 0, 3.0)
        flavor_four = self.get_flavor(4, "1024 MB Server", 3, 1024, 10, 10,
                                      2.0)

        return [flavor_one, flavor_two, flavor_three, flavor_four]

    def get_host(self, host_name, service, zone):
        h = mock.MagicMock()
        h.host_name = host_name
        h.service = service
        h.zone = zone
        return h

    def get_host_list(self):
        h_one = self.get_host('host1', 'nova-compute', 'nova1')
        h_two = self.get_host('host2', 'nova-cert', 'nova1')

        return [h_one, h_two]

    def get_service(self, id, binary, host, zone, status, state,
                    updated_at, disabled_reason):
        s = mock.MagicMock()
        s.id = id
        s.binary = binary
        s.host = host
        s.zone = zone
        s.status = status
        s.state = state
        s.updated_at = updated_at
        s.disabled_reason = disabled_reason

        return s

    def get_service_list(self):
        service_one = self.get_service(1, 'nova-compute', 'nova',
                                       'nova1', 'enabled', 'up',
                                       '2015-07-28T08:28:37.000000', None)
        service_two = self.get_service(2, 'nova-schedule', 'nova',
                                       'nova1', 'disabled', 'up',
                                       '2015-07-28T08:28:38.000000',
                                       'daily maintenance')

        return [service_one, service_two]

    def get_availability_zone(self, name, state):
        zone = mock.MagicMock()
        zone.zoneName = name
        zone.zoneState = state
        return zone

    def get_zone_list(self):
        zone_one = self.get_availability_zone('AZ1', 'available')
        zone_two = self.get_availability_zone('AZ2', 'not available')

        return [zone_one, zone_two]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import heatclient.v1.client as heatclient
from keystoneauth1.identity import v2
from keystoneauth1 import session
import keystoneclient.v2_0.client as ksclient
from oslo_log import log as logging

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return HeatV1Driver(name, keys, inbox, datapath, args)


class HeatV1Driver(datasource_driver.PollingDataSourceDriver,
                   datasource_driver.ExecutionDriver):

    STACKS = "stacks"
    STACKS_LINKS = "stacks_links"
    DEPLOYMENTS = "deployments"
    DEPLOYMENT_OUTPUT_VALUES = "deployment_output_values"
    RESOURCES = "resources"
    RESOURCES_LINKS = "resources_links"
    EVENTS = "events"
    EVENTS_LINKS = "events_links"

    # TODO(thinrichs): add snapshots
    value_trans = {'translation-type': 'VALUE'}
    stacks_links_translator = {
        'translation-type': 'HDICT',
        'table-name': STACKS_LINKS,
        'parent-key': 'id',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'href', 'translator': value_trans},
             {'fieldname': 'rel', 'translator': value_trans})}

    stacks_translator = {
        'translation-type': 'HDICT',
        'table-name': STACKS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
        ({'fieldname': 'id', 'translator': value_trans},
         {'fieldname': 'stack_name', 'translator': value_trans},
         {'fieldname': 'description', 'translator': value_trans},
         {'fieldname': 'creation_time', 'translator': value_trans},
         {'fieldname': 'updated_time', 'translator': value_trans},
         {'fieldname': 'stack_status', 'translator': value_trans},
         {'fieldname': 'stack_status_reason', 'translator': value_trans},
         {'fieldname': 'stack_owner', 'translator': value_trans},
         {'fieldname': 'parent', 'translator': value_trans},
         {'fieldname': 'links', 'translator': stacks_links_translator})}

    deployments_output_values_translator = {
        'translation-type': 'HDICT',
        'table-name': DEPLOYMENT_OUTPUT_VALUES,
        'parent-key': 'id',
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'deploy_stdout', 'translator': value_trans},
             {'fieldname': 'deploy_stderr', 'translator': value_trans},
             {'fieldname': 'deploy_status_code', 'translator': value_trans},
             {'fieldname': 'result', 'translator': value_trans})}

    software_deployment_translator = {
        'translation-type': 'HDICT',
        'table-name': DEPLOYMENTS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
        ({'fieldname': 'status', 'translator': value_trans},
         {'fieldname': 'server_id', 'translator': value_trans},
         {'fieldname': 'config_id', 'translator': value_trans},
         {'fieldname': 'action', 'translator': value_trans},
         {'fieldname': 'status_reason', 'translator': value_trans},
         {'fieldname': 'id', 'translator': value_trans},
         {'fieldname': 'output_values',
          'translator': deployments_output_values_translator})}

    resources_links_translator = {
        'translation-type': 'HDICT',
        'table-name': RESOURCES_LINKS,
        'parent-key': 'physical_resource_id',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'href', 'translator': value_trans},
             {'fieldname': 'rel', 'translator': value_trans})}

    resources_translator = {
        'translation-type': 'HDICT',
        'table-name': RESOURCES,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
        ({'fieldname': 'physical_resource_id', 'translator': value_trans},
         {'fieldname': 'logical_resource_id', 'translator': value_trans},
         {'fieldname': 'stack_id', 'translator': value_trans},
         {'fieldname': 'resource_name', 'translator': value_trans},
         {'fieldname': 'resource_type', 'translator': value_trans},
         {'fieldname': 'creation_time', 'translator': value_trans},
         {'fieldname': 'updated_time', 'translator': value_trans},
         {'fieldname': 'resource_status', 'translator': value_trans},
         {'fieldname': 'resource_status_reason', 'translator': value_trans},
         {'fieldname': 'links', 'translator': resources_links_translator})}

    events_links_translator = {
        'translation-type': 'HDICT',
        'table-name': EVENTS_LINKS,
        'parent-key': 'id',
        'selector-type': 'DICT_SELECTOR',
        'in-list': True,
        'field-translators':
            ({'fieldname': 'href', 'translator': value_trans},
             {'fieldname': 'rel', 'translator': value_trans})}

    events_translator = {
        'translation-type': 'HDICT',
        'table-name': EVENTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
        ({'fieldname': 'id', 'translator': value_trans},
         {'fieldname': 'physical_resource_id', 'translator': value_trans},
         {'fieldname': 'logical_resource_id', 'translator': value_trans},
         {'fieldname': 'stack_id', 'translator': value_trans},
         {'fieldname': 'resource_name', 'translator': value_trans},
         {'fieldname': 'event_time', 'translator': value_trans},
         {'fieldname': 'resource_status', 'translator': value_trans},
         {'fieldname': 'resource_status_reason', 'translator': value_trans},
         {'fieldname': 'links', 'translator': events_links_translator})}

    TRANSLATORS = [stacks_translator, software_deployment_translator,
                   resources_translator, events_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(HeatV1Driver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        auth = v2.Password(auth_url=self.creds['auth_url'],
                           username=self.creds['username'],
                           password=self.creds['password'],
                           tenant_name=self.creds['tenant_name'])
        sess = session.Session(auth=auth)
        keystone = ksclient.Client(**self.creds)
        endpoint = keystone.service_catalog.url_for(
            service_type='orchestration', endpoint_type='publicURL')
        self.heat = heatclient.Client(session=sess, endpoint=endpoint)
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'heat'
        result['description'] = ('Datasource driver that interfaces with'
                                 ' OpenStack orchestration aka heat.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        """Called when it is time to pull new data from this datasource."""
        LOG.debug("Grabbing Heat Stacks")
        stacks = self.heat.stacks.list()
        self._translate_stacks({'stacks': stacks})

        LOG.debug("Grabbing resources")
        stacks = self.heat.stacks.list()
        resources = self._get_resources(stacks)
        self._translate_resources(resources)

        LOG.debug("Grabbing events")
        stacks = self.heat.stacks.list()
        events = self._get_events(stacks)
        self._translate_events(events)

        LOG.debug("Grabbing the software deployments")
        deployments = {'deployments': self.heat.software_deployments.list()}
        self._translate_software_deployment(deployments)

    def _get_resources(self, stacks):
        rval = []
        for stack in stacks:
            resources = self.heat.resources.list(stack.id)
            for resource in resources:
                resource = resource.to_dict()
                resource['stack_id'] = stack.id
                rval.append(resource)
        return {'resources': rval}

    def _get_events(self, stacks):
        rval = []
        for stack in stacks:
            events = self.heat.events.list(stack.id)
            for event in events:
                event = event.to_dict()
                event['stack_id'] = stack.id
                rval.append(event)
        return {'events': rval}

    @ds_utils.update_state_on_changed(STACKS)
    def _translate_stacks(self, obj):
        """Translate the stacks represented by OBJ into tables."""
        LOG.debug("STACKS: %s", str(dict(obj)))
        row_data = HeatV1Driver.convert_objs(
            obj['stacks'], HeatV1Driver.stacks_translator)
        return row_data

    @ds_utils.update_state_on_changed(DEPLOYMENTS)
    def _translate_software_deployment(self, obj):
        """Translate the stacks represented by OBJ into tables."""
        LOG.debug("Software Deployments: %s", str(dict(obj)))
        row_data = HeatV1Driver.convert_objs(
            obj['deployments'], HeatV1Driver.software_deployment_translator)
        return row_data

    @ds_utils.update_state_on_changed(RESOURCES)
    def _translate_resources(self, obj):
        """Translate the resources represented by OBJ into tables."""
        LOG.debug("Resources: %s", str(dict(obj)))
        row_data = HeatV1Driver.convert_objs(
            obj['resources'], HeatV1Driver.resources_translator)
        return row_data

    @ds_utils.update_state_on_changed(EVENTS)
    def _translate_events(self, obj):
        """Translate the events represented by OBJ into tables."""
        LOG.debug("Events: %s", str(dict(obj)))
        row_data = HeatV1Driver.convert_objs(
            obj['events'], HeatV1Driver.events_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.heat, action, action_args)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import contextlib
import json
import shutil
import tempfile

from oslo_config import cfg
from oslo_log import log as logging
import six

LOG = logging.getLogger(__name__)

utils_opts = [
    cfg.StrOpt('tempdir',
               help='Explicitly specify the temporary working directory'),
]
CONF = cfg.CONF
CONF.register_opts(utils_opts)


@contextlib.contextmanager
def tempdir(**kwargs):
    argdict = kwargs.copy()
    if 'dir' not in argdict:
        argdict['dir'] = CONF.tempdir
    tmpdir = tempfile.mkdtemp(**argdict)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.error(_('Could not remove tmpdir: %s'), e)


def value_to_congress(value):
    if isinstance(value, six.string_types):
        # TODO(ayip): This throws away high unicode data because congress does
        # not have full support for unicode yet.  We'll need to fix this to
        # handle unicode coming from datasources.
        try:
            six.text_type(value).encode('ascii')
        except UnicodeEncodeError:
            LOG.warning('Ignoring non-ascii characters')
        # Py3: decode back into str for compat (bytes != str)
        return six.text_type(value).encode('ascii', 'ignore').decode('ascii')
    # Check for bool before int, because True and False are also ints.
    elif isinstance(value, bool):
        return str(value)
    elif (isinstance(value, six.integer_types) or
          isinstance(value, float)):
        return value
    return str(value)


class Location (object):
    """A location in the program source code."""

    __slots__ = ['line', 'col']

    def __init__(self, line=None, col=None, obj=None):
        try:
            self.line = obj.location.line
            self.col = obj.location.col
        except AttributeError:
            pass
        self.col = col
        self.line = line

    def __str__(self):
        s = ""
        if self.line is not None:
            s += " line: {}".format(self.line)
        if self.col is not None:
            s += " col: {}".format(self.col)
        return s

    def __repr__(self):
        return "Location(line={}, col={})".format(
            repr(self.line), repr(self.col))

    def __hash__(self):
        return hash(('Location', hash(self.line), hash(self.col)))


def pretty_json(data):
    print (json.dumps(data, sort_keys=True,
                      indent=4, separators=(',', ': ')))
